// txtmus - cmdline.js - command line
//
// Copyright (C) 2014-2022 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.

// mus object
    var	mus,				// (global for 'toxxx.js')

// user definitions
      user = {
	read_file: function(fn) {	// read a file (main or included)
	    var	i,
		p = fn,
		file = tm.readFile(p)

		if (!file && fn[0] != '/') {
			for (i = 0; i < tm.path.length; i++) {
				p = tm.path[i] + '/' + fn
				file = tm.readFile(p)
				if (file)
					break
			}
		}

		if (!file)
			return file

		// memorize the file path
		i = p.lastIndexOf('/')
		if (i > 0) {
			p = p.slice(0, i)
			if (tm.path.indexOf(p) < 0)
				tm.path.unshift(p)
		}

		// convert the file content into a Unix string
		i = file.indexOf('\r')
		if (i >= 0)
			file = file.replace(/\r\n|\r/g, '\n')

		// load the required modules (synchronous)
		tm.modules.load(file)

		return file
	},
	errtxt: '',
	errmsg:			// print or store the error messages
		typeof tm.printErr == 'function'
			? function(msg, l, c) { tm.printErr(msg) }
			: function(msg, l, c) { user.errtxt += msg + '\n' }
} // user

if (!tm.path)
	tm.path = []		// path to ABC files - from env TMPATH

// treat a file
function do_file(fn) {
    var	mt,
	mb = user.read_file(fn)

	if (!mb) {
		if (fn.slice(0, 7) != "default")
			user.errmsg("Cannot read file '" + fn + "'")
		return
	}

	// if the source is not ABC, switch the entry
	switch (fn.split('.').slice(-1)[0]) {
	case "alda":
		tm.modules.load("%%alda 1\n")
		mus.tosvg("cmd", "%%alda 1\n")
		break
	}

	// generate
	try {
		mus.tosvg(fn, mb)
	} catch (e) {
		tm.abort(e)
	}
} // do_file()

function tmcmd(cmd, args, interp_name) {
    var	arg, parm, fn

	tm.abort = function(e) {
		tm.printErr('javascript error: ' + e.message +
			'\nStack:\n'  + e.stack)
		tm.quit()
	} // abort()

	// put the last options before the last file
	function arg_reorder(a) {
	    var	f,
		i = a.length - 2

		while (i > 2 && a[i].slice(0, 2) == '--')
			i -= 2
		f = a[--i]
		a.splice(i, 1)
		a.push(f)
	} // arg_reorder()

	// load the preferences
	function pref() {
	    var	i, j,
		mb = tm.readFile("pref.js")

		if (!mb)
			return
		i = mb.indexOf('%%')
		if (i < 0)
			return
		j = mb.lastIndexOf('%%')
		j = mb.indexOf('\n', j)
		mb = mb.slice(i, j)
		if (mb.indexOf('\\n') > 0)
			mb = mb.replace(/\\n\\/g, '')
		tm.modules.load(mb)
		mus.tosvg(cmd, mb)
	} // pref()

	// if the first argument is a javascript file, load it
	if (args[0] && args[0].slice(-3) == '.js') {
		tm.loadjs(args[0])
		args.shift()
	}

	if (!args[0]) {
		tm.printErr('music translator with ' + interp_name +
			' and txtmus-' + tm.version + ' ' +
					tm.vdate +
			'\nUsage:\n  ' + cmd +
		    ' [script.js] [options] txt_file [[options] txt_file]* [options]\n\
Arguments:\n\
  script.js  backend script - default: tohtml.js (HTML+SVG)\n\
  options    txt options (the last options are moved before the last file)\n\
  txt_file   text music file')
		tm.quit()
	}

	// the default output is HTML+SVG
	if (typeof tm.mus_init != 'function')
		tm.loadjs("tohtml.js")

	// initialize the backend
	mus = new tm.Mus(user)
	tm.mus_init(args)

	pref()				// load the preferences

	// put the last options before the last txt file
	if (args.length > 2 && args[args.length - 2].slice(0, 2) == '--')
		arg_reorder(args)

	while (1) {
		arg = args.shift()
		if (!arg)
			break
		if (arg[0] == "-" && arg[1] == "-") {
			parm = arg.replace('--', 'I:')
				+ " " + args.shift() + "\n"
			tm.modules.load(parm)
			mus.tosvg(cmd, parm)
		} else {
			do_file(arg)
			// keep the selection
		}
	}
	tm.mus_end()
} // tmcmd()
