// txtmus - deco.js - decorations
//
// Copyright (C) 2014-2024 Jean-Francois Moine
//
// This file is part of txtmus-core.
//
// txtmus-core is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus-core is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus-core.  If not, see <http://www.gnu.org/licenses/>.

// Decoration objects
// dd {			// decoration definition (static)
//	dd_en,			// definition of the ending decoration
//	dd_st,			// definition of the starting decoration
//	func,			// function
//	glyph,			// glyph
//	h,			// height / ascent
//	hd,			// descent
//	inv,			// inverted glyph
//	name,			// name
//	str,			// string
//	wl,			// left width
//	wr,			// right width
// }
// de {			// decoration elements (in an array - one list per music line)
//	dd,			// definition of the decoration
//	defl {			// flags
//		noen,			// no end of this decoration
//		nost,			// no start of this decoration
//	},
//	ix,			// index of the decoration in the 'de' list
//	lden,			// end of a long decoration
//	ldst,			// start of a long decoration if true
//	m,			// note index when note decoration
//	prev,			// previous decoration (hack for 'tr~~~~~')
//	s,			// symbol
//	start,			// start of the decoration (in the ending element)
//	up,			// above the symbol
//	val,			// value
//	x,			// x offset
//	y,			// y offset
// }

    var	dd_tb = {},		// definition of the decorations
	a_de,			// array of the decoration elements
	cross,			// cross voice decorations

    decos = {		// decorations - populate with standard decorations
	dot: "0 stc 6 1.5 1",
	tenuto: "0 emb 6 4 3",
	wedge: "0 wedge 8 1.5 1",		// (staccatissimo or spiccato)
	slide: "1 sld 3 7 1",
	arpeggio: "2 arp 12 10 3",
	downbow: "3 dnb 9 4 6",
	gmark: "3 grm 7 4 6",
	invertedturn: "3 turn 6,1 5 7",
	lowermordent: "3 lmrd 6,5 4 6",
	marcato: "3 marcato 9 5 5",
	mordent: "3 lmrd 6,5 4 6",
	open: "3 opend 8 3 3",
	plus: "3 dplus 8,2 2 4",
	roll: "3 roll 5,4 5 6",
	snap: "3 snap 10 3 3",
	thumb: "3 thumb 10 3 3",
	trill: "3 trl 14 5 8",
	turn: "3 turn 7,2.5 5 6",
	turnx: "3 turnx 7,2.5 5 6",
	upbow: "3 upb 10,2 3 7",
	uppermordent: "3 umrd 6,5 4 6",
	"0": "3 fng 5,5 3 3 0",
	"1": "3 fng 5,5 3 3 1",
	"2": "3 fng 5,5 3 3 2",
	"3": "3 fng 5,5 3 3 3",
	"4": "3 fng 5,5 3 3 4",
	"5": "3 fng 5,5 3 3 5",
	"+": "3 dplus 8,2 2 4",
	"^": "3 marcato 9 5 5",
	"8va(": "5 8va 12 6 6",
	"8va)": "5 8va 12 6 6",
	"8vb(": "4 8vb 10,5 6 6",
	"8vb)": "4 8vb 10,5 6 6",
	"15ma(": "5 15ma 12 9 9",
	"15ma)": "5 15ma 12 9 9",
	"15mb(": "4 15mb 12 9 9",
	"15mb)": "4 15mb 12 9 9",
	">": "5 accent 3.5,3.5 4 4",
	accent: "5 accent 3.5,3.5 4 4",
	breath: "5 brth 0 1 16",
	caesura: "5 caes 0 1 20",
	coda: "5 coda 22,5 10 10",
	dacapo: "5 dacs 16 20 20 Da Capo",
	dacoda: "5 dacs 16 20 20 Da Coda",
	"D.C.": "5 dcap 16,3 12 12",
	"D.S.": "5 dsgn 16,3 12 12",
	"D.C.alcoda": "5 dacs 16 38 38 D.C. al Coda",
	"D.S.alcoda": "5 dacs 16 38 38 D.S. al Coda",
	"D.C.alfine": "5 dacs 16 38 38 D.C. al Fine",
	"D.S.alfine": "5 dacs 16 38 38 D.S. al Fine",
	emphasis: "5 accent 3.5,3.5 4 4",
	fermata: "5 hld 12 7.5 7.5",
	fine: "5 dacs 16 14 14 Fine",
	segno: "5 sgno 22,2 5 5",
	"trill(": "5 ltr 8 0 0",
	"trill)": "5 ltr 8 0 0",
	f: "6 f 12,5 3 4",
	ff: "6 ff 12,5 8 5",
	fff: "6 fff 12,5 11 9",
	ffff: "6 ffff 12,5 15 12",
	mf: "6 mf 12,5 8 10",
	mp: "6 mp 12,5 9 10",
	p: "6 p 12,5 3 6",
	pp: "6 pp 12,5 8 9",
	ppp: "6 ppp 12,5 14 11",
	pppp: "6 pppp 12,5 14 17",
	pralltriller: "3 umrd 6,5 4 6",
	sfz: "6 sfz 12,5 9 9",
	"crescendo(": "6 cresc 15,2 0 0",
	"crescendo)": "6 cresc 15,2 0 0",
	"<(": "6 cresc 15,2 0 0",
	"<)": "6 cresc 15,2 0 0",
	"diminuendo(": "6 dim 15,2 0 0",
	"diminuendo)": "6 dim 15,2 0 0",
	">(": "6 dim 15,2 0 0",
	">)": "6 dim 15,2 0 0",
	invertedfermata: "7 hld 12 8 8",
	ped: "6 ped 9,2 6 10",
	"ped-up": "6 pedoff 9,2 4 4",
	"ped(": "7 lped 14 1 1",
	"ped)": "7 lped 14 1 1",
	"-(": "8 gliss 0 0 0",
	"-)": "8 gliss 0 0 0",
	"~(": "8 glisq 0 0 0",
	"~)": "8 glisq 0 0 0",
// internal
//	"head-xxx": "9 0 0 0 0",
//	color: "10 0 0 0 0",
	invisible: "32 0 0 0 0",
	beamon: "33 0 0 0 0",
	trem1: "34 0 0 0 0",
	trem2: "34 0 0 0 0",
	trem3: "34 0 0 0 0",
	trem4: "34 0 0 0 0",
	xstem: "35 0 0 0 0",
	beambr1: "36 0 0 0 0",
	beambr2: "36 0 0 0 0",
	rbstop: "37 0 0 0 0",
	"/": "38 0 0 6 6",
	"//": "38 0 0 6 6",
	"///": "38 0 0 6 6",
	"beam-accel": "39 0 0 0 0",
	"beam-rall": "39 0 0 0 0",
	stemless: "40 0 0 0 0",
	rbend: "41 0 0 0 0",
	editorial: "42 0 0 0 0",
	"sacc-1": "3 sacc-1 6,4 4 4",
	sacc3: "3 sacc3 6,5 4 4",
	sacc1: "3 sacc1 6,4 4 4",
	courtesy: "43 0 0 0 0",
	"cacc-1": "3 cacc-1 0 0 0",
	cacc3: "3 cacc3 0 0 0",
	cacc1: "3 cacc1 0 0 0",
	"tie(": "44 0 0 0 0",
	"tie)": "44 0 0 0 0",
	longphrase: "45 l 0 0 0",		// special bars
	mediumphrase: "45 m 0 0 0",
	short: "45 s 0 0 0",
	shortphrase: "45 t 0 0 0",
	tick: "45 t 0 0 0"},

	// types of decoration per function
	f_near = [
		d_near,		// 0 - near the note
		d_slide,	// 1
		d_arp		// 2
	],
	f_note = [
		null, null, null,
		d_upstaff,	// 3 - tied to note
		d_upstaff	// 4 (below the staff)
	],
	f_staff = [
		null, null, null, null, null,
		d_upstaff,	// 5 (above the staff)
		d_upstaff,	// 6 - tied to staff (dynamic marks)
		d_upstaff	// 7 (below the staff)
	]

// -- get the max/min vertical offset --
function y_get(st, up, x, w) {
    var	y,
	p_staff = staff_tb[st],
	i = (x / 3) | 0,		// !! / realwidth * YSTEP !!
	j = ((x + w) / 3) | 0

	if (i < 0)
		i = 0
	if (j >= YSTEP) {
		j = YSTEP - 1
		if (i > j)
			i = j
	}
	if (up) {
		y = p_staff.top[i++]
		while (i <= j) {
			if (y < p_staff.top[i])
				y = p_staff.top[i]
			i++
		}
	} else {
		y = p_staff.bot[i++]
		while (i <= j) {
			if (y > p_staff.bot[i])
				y = p_staff.bot[i]
			i++
		}
	}
	return y
} // y_get()

// -- adjust the vertical offsets --
function y_set(st, up, x, w, y) {
    var	p_staff = staff_tb[st],
	i = (x / 3) | 0,		// !! / realwidth * YSTEP !!
	j = ((x + w) / 3) | 0

	// (may occur when annotation on 'y' at start of an empty staff)
	if (i < 0)
		i = 0
	if (j >= YSTEP) {
		j = YSTEP - 1
		if (i > j)
			i = j
	}
	if (up) {
		while (i <= j) {
			if (p_staff.top[i] < y)
				p_staff.top[i] = y
			i++
		}
	} else {
		while (i <= j) {
			if (p_staff.bot[i] > y)
				p_staff.bot[i] = y
			i++
		}
	}
} // y_set()

// get the staff position
// - of the ornements
function up3(s, pos) {
	switch (pos & 0x07) {
	case C.SL_ABOVE:
		return 1 // true
	case C.SL_BELOW:
		return 0 // false
	}
	return !s.second
} // up3()

// - of the dynamic and volume marks
function up6(s, pos) {
	switch (pos & 0x07) {
	case C.SL_ABOVE:
		return 1 // true
	case C.SL_BELOW:
		return 0 // false
	}
	if (s.multi > 0)
		return 1 // true
	if (s.second
	 || !s.p_v.have_ly)
		return 0 // false

	// above if the lyrics are below the staff
	return (s.pos.voc & 0x07) != C.SL_ABOVE
} // up6()

// -- drawing functions --
// 2: special case for arpeggio
function d_arp(de) {
    var	m, h, dx,
	s = de.s,
	dd = de.dd,
	xc = dd.wr

	if (s.typ == C.note) {
		for (m = 0; m <= s.nhd; m++) {
			if (s.notes[m].acc) {
				dx = s.notes[m].shac
			} else {
				dx = 1 - s.notes[m].shhd
				switch (s.head) {
				case C.SQUARE:
					dx += 3.5
					break
				case C.OVALBARS:
				case C.OVAL:
					dx += 2
					break
				}
			}
			if (dx > xc)
				xc = dx
		}
	}
	h = 3 * (s.notes[s.nhd].pit - s.notes[0].pit) + 4
	m = dd.h			// minimum height
	if (h < m)
		h = m

	de.val = h
	de.x -= xc
	de.y = 3 * (s.notes[0].pit + s.notes[s.nhd].pit) / 2 - h / 2 - 3
} // d_arp()

// 0: near the note (dot, tenuto)
function d_near(de) {
    var	y,
	up = de.up,
	s = de.s,
	dd = de.dd

	y = up ? s.ymx : s.ymn
	if (y > 0 && y < 24) {
		y = (((y + 9) / 6) | 0) * 6 - 6	// between lines
	}
	if (up) {
		y += dd.hd
		s.ymx = y + dd.h
	} else if (dd.name[0] == 'w') {		// wedge (no descent)
		de.inv = 1 //true
		y -= dd.h
		s.ymn = y
	} else {
		y -= dd.h
		s.ymn = y - dd.hd
	}
	de.x -= dd.wl
	de.y = y
	if (s.typ == C.note)
		de.x += s.notes[s.stem >= 0 ? 0 : s.nhd].shhd
	if (dd.name[0] == 'd') {		// if dot (staccato)
		if (!(s.beam_st && s.beam_end)) { // if in a beam sequence
			if (up) {
				if (s.stem > 0)
					de.x += 3.5	// stem_xoff
			} else {
				if (s.stem < 0)
					de.x -= 3.5
			}
		} else {
			if (up && s.stem > 0) {
				y = s.y + (y - s.y) * .6
				if (y >= 27) {
					de.y = y	// put the dot a bit lower
					s.ymx = de.y + dd.h
				}
			}
		}
	}
} // d_near()

// 1: special case for slide
function d_slide(de) {
	var	m, dx,
		s = de.s,
		yc = s.notes[0].pit,
		xc = 5

	for (m = 0; m <= s.nhd; m++) {
		if (s.notes[m].acc) {
			dx = 4 + s.notes[m].shac
		} else {
			dx = 5 - s.notes[m].shhd
			switch (s.head) {
			case C.SQUARE:
				dx += 3.5
				break
			case C.OVALBARS:
			case C.OVAL:
				dx += 2
				break
			}
		}
		if (s.notes[m].pit <= yc + 3 && dx > xc)
			xc = dx
	}
//	de.x = s.x - xc
	de.x -= xc
	de.y = 3 * yc
} // d_slide()

// special case for long decoration
function d_long(de) {
	if (de.ldst)
		return
    var	y, w, tmp,
	up = de.start.up,
	dd = de.dd,
	de2 = de.prev,			// decoration at the same offset
	s2 = de.s,
	st = s2.st,
	s = de.start.s,			// start of the long decoration
	x = s.x

	// shift the starting point of a long decoration
	// in the cases "T!trill(!" and "!pp!!<(!"
	// (side effect on x)
	function sh_st() {
	    var	de3,
		de2 = de.start,			// start of the decoration
		s = de2.s,
		i = de2.ix			// index of the current decoration

		while (--i >= 0) {
			de3 = a_de[i]
			if (!de3 || de3.s != s)
				break
		}
		while (1) {			// loop on the decorations of the symbol
			i++
			de3 = a_de[i]
			if (!de3 || de3.s != s)
				break
			if (de3 == de2)
				continue
			if (!(up ^ de3.up)
			 && (de3.dd.name == "trill"
			  || de3.dd.func == 6)) {	// dynamic
				x += de3.dd.wr + 3
				break
			}
		}
	} // sh_st()

	// shift the ending point of a long decoration
	// (side effect on w)
	function sh_en() {
	    var	de3,
		i = de.ix			// index of the current decoration

		while (--i > 0) {
			de3 = a_de[i]
			if (!de3 || de3.s != s2)
				break
		}
		while (1) {			// loop on the decorations of the symbol
			i++
			de3 = a_de[i]
			if (!de3 || de3.s != s2)
				break
//			if (de3 == de || de3 == de2)
			if (de3 == de)
				continue
			if (!(up ^ de3.up)
			 && de3.dd.func == 6) {	// if dynamic mark
				w -= de3.dd.wl
				break
			}
		}
	} //sh_en()

	// start of d_long()
	if (de2) {			// same height
		x = de2.s.x + de.dd.wl + 2
		de2.val -= de2.dd.wr
		if (de2.val < 8)
			de2.val = 8
	}
	de.st = st
	de.up = up

	sh_st()				// shift the starting point?

	if (de.defl.noen) {		// if no decoration end
		w = de.x - x
		if (w < 20) {
			x = de.x - 20 - 3
			w = 20
		}
	} else {
		w = s2.x - x - 4
		sh_en(de)		// shift the ending point?
		if (w < 20)
			w = 20
	}
	y = y_get(st, up, x - dd.wl, w)
	if (up) {
		tmp = staff_tb[s.st].topbar + 2
		if (y < tmp)
			y = tmp
	} else {
		tmp = staff_tb[s.st].botbar - 2
		if (y > tmp)
			y = tmp
		y -= dd.h
	}
	if (de2) {			// if same height
		if (up) {
			if (y < de2.y)
				y = de2.y	// (only on one note)
		} else {
			if (y >= de2.y) {
				y = de2.y
			} else {
				do {
					de2.y = y
					de2 = de2.prev	// go backwards
				} while (de2)
			}
		}
	}

	de.lden = 0 //false
	de.val = w
	de.x = x
	de.y = y
	if (up)
		y += dd.h
	else
		y -= dd.hd
	y_set(st, up, x, w, y)
	if (up)
		s.ymx = s2.ymx = y
	else
		s.ymn = s2.ymn = y
} // d_long()

// 3, 4, 5, 7: above (or below) the staff
function d_upstaff(de) {

	// don't treat here the long decorations
	if (de.ldst)			// if long deco start
		return
	if (de.start) {			// if long decoration
		d_long(de)
		return
	}

    var	y, inv,
	up = de.up,
	s = de.s,
	dd = de.dd,
	x = de.x,
	w = dd.wl + dd.wr

	// glyphs inside the staff
	switch (dd.glyph) {
	case "brth":
	case "caes":
		y = staff_tb[s.st].topbar + 2 + dd.hd
		if (y < s.ymx)
			y = s.ymx
		for (s = s.ts_next; s; s = s.ts_next)
			if (s.seqst)
				break
		x += ((s ? s.x : realwidth) - x) * .45
		de.x = x
		de.y = y
		return
	}

	if (s.nhd)
		x += s.notes[s.stem >= 0 ? 0 : s.nhd].shhd

	switch (dd.ty) {
	case '@':
	case '<':
	case '>':
		y = de.y
		break
	}
	if (y == undefined) {
		if (up) {
			y = y_get(s.st, up, x - dd.wl, w) + dd.hd
			if (de.y > y)
				y = de.y
			s.ymx = y + dd.h
		} else {
			y = y_get(s.st, up, x - dd.wl, w) - dd.h
			if (de.y < y)
				y = de.y
			if (dd.name == "fermata"
			 || dd.glyph == "accent"
			 || dd.glyph == "roll")
				de.inv = 1 //true
			s.ymn = y - dd.hd
		}
	}

	if (dd.wr > 10 && x > realwidth - dd.wr)
		de.x = x = realwidth - w

//    if (dd.func == 6
//     && ((de.pos & C.SL_ALI_MSK) == C.SL_ALIGN
//      || ((de.pos & C.SL_ALI_MSK) == 0
//       && de.s.fmt.dynalign > 0)))	// if align
//	;
//    else
	if (up)
		y_set(s.st, 1, x - dd.wl, w, y + dd.h)
	else
		y_set(s.st, 0, x - dd.wl, w, y - dd.hd)
	de.y = y
} // d_upstaff()

// add a decoration
// syntax:
//	%%deco <name> <c_func> <glyph> <h> <wl> <wr> [<str>]
// "<h>" may be followed by ",<hd>" (descent)
function deco_add(parm) {
    var	dv = parm.match(/(\S*)\s+(.*)/)

	decos[dv[1]] = dv[2]
}

// define a decoration
// nm is the name of the decoration
// nmd is the name of the definition in the table 'decos'
function deco_def(nm, nmd) {
	if (!nmd)
		nmd = nm
    var a, dd, dd2, nm2, c, i, elts, str, hd,
	text = decos[nmd]

	// check if a long decoration with number
	if (!text && /\d[()]$/.test(nmd))
		text = decos[nmd.replace(/\d/, '')]
//fixme: if any digit in the decoration name, use
//		text = decos[nmd.replace(/(\d)([()])$/,'$2')]

	if (!text) {
		if (cfmt.decoerr)
			error(1, null, "Unknown decoration '$1'", nm)
		return //undefined
	}

	// extract the values
	a = text.match(/(\d+)\s+(.+?)\s+([0-9.,]+)\s+([0-9.]+)\s+([0-9.]+)/)
	if (!a) {
		error(1, null, "Invalid decoration '$1'", nm)
		return //undefined
	}
    var	func = +a[1],
//	glyph = a[2],
	h = a[3],
	wl = +a[4],
	wr = +a[5]

	if (isNaN(func)) {
		error(1, null, "%%deco: bad C function value '$1'", a[1])
		return //undefined
	}
	if (func > 10
	 && (func < 32 || func > 45)) {
		error(1, null, "%%deco: bad C function index '$1'", func)
		return //undefined
	}

	if (h.indexOf(',') > 0) {
		h = h.split(',')
		hd = h[1]
		h = h[0]
	} else {
		hd = 0
	}
	if (h > 50 || wl > 80 || wr > 80) {
		error(1, null, "%%deco: abnormal h/wl/wr value '$1'", text)
		return //undefined
	}

	// create/redefine the decoration
	dd = dd_tb[nm]
	if (!dd) {
		dd = {
			name: nm
		}
		dd_tb[nm] = dd
	}

	// set the values
	dd.func = nm.indexOf("head-") == 0 ? 9 : func
	dd.glyph = a[2]
	dd.h = +h
	dd.hd = +hd
	dd.wl = wl
	dd.wr = wr
	str = text.replace(a[0], '').trim()
	if (str) {				// optional string
		if (str[0] == '"')
			str = str.slice(1, -1)
		if (str[0] == '@') {
			c = str.match(/^@([0-9.-]+),([0-9.-]+)[ ;]?/)
			if (!c) {
				error(1, null, "%%deco: bad position '$1'", str)
				return
			}
			dd.dx = +c[1]		// x and y offsets
			dd.dy = +c[2]
			str = str.replace(c[0], '')
		}
		dd.str = str
	}

	// compatibility
	if (dd.func == 6 && dd.str == undefined)
		dd.str = nm

	// link the start and end of long decorations
	c = nm.slice(-1)
	if (c == '('
	 || (c == ')' && nm.indexOf('(') < 0)) {	// not (#)
		dd.str = null				// (no string)
		nm2 = nm.slice(0, -1) + (c == '(' ? ')' : '(')
		dd2 = dd_tb[nm2]
		if (dd2) {
			if (c == '(') {
				dd.dd_en = dd2
				dd2.dd_st = dd
			} else {
				dd.dd_st = dd2
				dd2.dd_en = dd
			}
		}
	}
	return dd
} // deco_def()

// define a cross-voice tie
// @nm = decoration name
// @s = note symbol
// @nt1 = note
function do_ctie(nm, s, nt1) {
    var	nt2 = cross[nm],
	nm2 = nm.slice(0, -1) + (nm.slice(-1) == '(' ? ')' : '(')

	if (nt2) {
		error(1, s, "Conflict on !$1!", nm)
		return
	}
	nt1.s = s
	nt2 = cross[nm2]
	if (!nt2) {
		cross[nm] = nt1		// keep the start/end
		return
	}
	if (nm.slice(-1) == ')') {
		nt2 = nt1
		nt1 = cross[nm2]
	}
	cross[nm2] = null
	if (nt1.midi != nt2.midi
	 || nt1.s.time + nt1.s.dur != nt2.s.time) {
		error(1, s, errs.bad_tie)
	} else {
		nt1.tie_ty = C.SL_AUTO
		nt1.tie_e = nt2
		nt2.tie_s = nt1
		nt1.s.ti1 = nt2.s.ti2 = 1 //true
	}
} // do_ctie()

// get/create the definition of a decoration
function get_dd(nm) {
    var	ty, p,
	dd = dd_tb[nm]

	if (dd)
		return dd
	if ("<>^_@".indexOf(nm[0]) >= 0	// if position
	 && !/^([>^]|[<>]\d?[()])$/.test(nm)) {
		ty = nm[0]
		if (ty == '@') {
			p = nm.match(/@([-\d]+),([-\d]+)/)
			if (!p) {
				error(1, s, "Bad position in !$1!", nm)
				return
			}
			ty = p[0]
		}
		dd = deco_def(nm, nm.replace(ty, ''))
	} else {
		dd = deco_def(nm)
	}
	if (!dd)
		return
	if (ty) {
		if (ty[0] == '@') {		// if with x,y
			dd.x = +p[1]
			dd.y = +p[2]
			ty = '@'
		}
		dd.ty = ty
	}
	return dd
} // get_dd()

// -- convert the decorations --
function deco_cnv(s, prev) {
    var	i, j, dd, nm, note, court

	while (1) {
		nm = a_dcn.shift()
		if (!nm)
			break
		dd = get_dd(nm)
		if (!dd)
			continue

		// special decorations
		switch (dd.func) {
		case 0:			// near
			if (s.bar_type && nm == "dot") {
				s.bar_dotted = 1//true
				continue
			}
			// fall thru
		case 1:			// slide
		case 2:			// arp
//			if (s.typ != C.note && s.typ != C.rest) {
			if (!s.notes) {
				error(1, s, errs.must_note_rest, nm)
				continue
			}
			break
		case 4:			// below the staff
		case 5:			// upstaff
			i = nm.match(/1?[85]([vm])([ab])([()])/)
			if (i) {				// if ottava
				j = i[1] == 'v' ? 1 : 2
				if (i[2] == 'b')
					j = -j
				if (!s.ottava)
					s.ottava = []
				s.ottava[i[3] == '(' ? 0 : 1] = j
				glovar.ottava = 1 //true
				break
			}
			i = ""
			switch (dd.glyph) { // put time marks
			case "dacs":	// dacapo, dacoda, D.C.../D.S..., fine
				switch (dd.str[2]) {
				case 'C':
					i = "dcap"	// D.C. [alcoda, alfine]
					break
				case 'S':
					i = "dsgn"	// D.S. [alcoda, alfine]
					break
				case 'n':
					i = "fine"
					break
				case 'c':
					i = dd.str[3] == 'a'
						? "dcap"	// dacapo
						: "coda"	// dacoda
					break
				}
				break
			case "coda":
			case "dcap":	// D.C.
			case "dsgn":	// D.S.
			case "sgno":
				i = dd.glyph
				break
			}
			if (i) {
				if (s.timark)
					error(0, s, "Second play mark !$1! ignored",
							dd.name)
				else
					s.timark = i	// flow mark
			}
			break
		case 8:			// gliss
			if (s.typ != C.note) {
				error(1, s, errs.must_note, nm)
				continue
			}
			note = s.notes[s.nhd] // move to the upper note of the chord
			if (!note.a_dd)
				note.a_dd = []
			note.a_dd.push(dd)
			continue
		case 9:			// alternate head
			if (!s.notes) {
				error(1, s, errs.must_note_rest, nm)
				continue
			}

			// move the alternate head of the chord to the notes
			for (j = 0; j <= s.nhd; j++) {
				note = s.notes[j]
				note.invis = 1 //true
				if (!note.a_dd)
					note.a_dd = []
				note.a_dd.push(dd)
			}
			continue
		case 10:		// color
			s.color = nm
			break
		case 32:		// invisible 
			s.invis = 1 //true
			break
		case 33:		// beamon
			s.beamon = 1 //true
			break
		case 34:		// trem1..trem4
			if (s.typ != C.note
			 || !prev
			 || prev.typ != C.note
			 || s.dur != prev.dur) {
				error(1, s,
					"!$1! must be on the last of a couple of notes",
					nm)
				continue
			}
			s.trem2 = 1 //true
			s.beam_end = 1 //true
			s.beam_st = 0 //false
			prev.beam_st = 1 //true
			prev.beam_end = 0 //false
			s.ntrem = prev.ntrem = +nm[4]
			for (j = 0; j <= s.nhd; j++)
				s.notes[j].dur *= 2
			for (j = 0; j <= prev.nhd; j++)
				prev.notes[j].dur *= 2
			break
		case 35:		// xstem
			if (s.typ != C.note) {
				error(1, s, errs.must_note, nm)
				continue
			}
			s.xstem = 1 //true
			break
		case 36:		// beambr1 / beambr2
			if (s.typ != C.note) {
				error(1, s, errs.must_note, nm)
				continue
			}
			s.bmbrk = +nm[6]
			break
		case 37:		// rbstop
			s.rbstop = 1	// open
			break
		case 38:		// /, // and /// = tremolo
			if (s.typ != C.note) {
				error(1, s, errs.must_note, nm)
				continue
			}
			s.trem1 = 1 //true
			s.ntrem = nm.length	// 1, 2 or 3
			break
		case 39:		// beam-accel/beam-rall
			if (s.typ != C.note) {
				error(1, s, errs.must_note, nm)
				continue
			}
			s.feathered_beam = nm[5] == 'a' ? 1 : -1
			break
		case 40:		// stemless
			s.stemless = 1 //true
			break
		case 41:		// rbend
			s.rbstop = 2	// with end
			break
		case 42:		// editorial
			if (s.typ != C.note) {
				error(1, s, errs.must_note, nm)
				continue
			}
			if (!s.notes[0].acc)
				continue
			nm = "sacc" + s.notes[0].acc.toString() // small accidental
			dd = dd_tb[nm]
			if (!dd) {
				dd = deco_def(nm)
				if (!dd) {
					error(1, s, errs.bad_val, "!editorial!")
					continue
				}
			}
			delete s.notes[0].acc
			curvoice.acc[s.notes[0].pit + 37] = 0	// ignore the accidental
			break
		case 43:		// courtesy
			if (s.typ != C.note) {
				error(1, s, errs.must_note, nm)
				continue
			}
			j = curvoice.acc[s.notes[0].pit + 37]
			if (s.notes[0].acc || !j)
				continue
			court = 1			// defer
			break
		case 44:		// cross-voice ties
			if (s.typ != C.note) {
				error(1, s, errs.must_note, nm)
				continue
			}
			do_ctie(nm, s, s.notes[0])	// (only one note for now)
			continue
		case 45:		// special bars
			if (!s.bar_type) {
				s = {
					typ: C.bar
				}
				sym_link(s)
			}
			s.bar_type = dd.glyph
			continue
		}

		// add the decoration in the symbol
		if (!s.a_dd)
			s.a_dd = []
		s.a_dd.push(dd)
	}

	// handle the possible courtesy accidental
	if (court) {
		a_dcn.push("cacc" + j)
		dh_cnv(s, s.notes[0])
	}
}

// -- convert head decorations --
// The decorations are in the global array a_dcn
function dh_cnv(s, nt) {
    var	k, nm, dd

	while (1) {
		nm = a_dcn.shift()
		if (!nm)
			break
		dd = get_dd(nm)
		if (!dd)
			continue

		switch (dd.func) {
		case 0:
		case 1:
		case 3:
		case 4:
		case 8:			// gliss
			break
		default:
//		case 2:			// arpeggio
//		case 5:			// trill
//		case 7:			// d_cresc
			error(1, s, "Cannot have !$1! on a head", nm)
			continue
		case 9:			// head replacement
			nt.invis = 1 //true
			break
		case 32:		// invisible
			nt.invis = 1 //true
			continue
		case 10:		// color
			nt.color = nm
			continue
		case 40:		// stemless chord (abcm2ps behaviour)
			s.stemless = 1 //true
			continue
		case 44:		// cross-voice ties
			do_ctie(nm, s, nt)
			continue
		}

		// add the decoration in the note
		if (!nt.a_dd)
			nt.a_dd = []
		nt.a_dd.push(dd)
	}
} // dh_cnv()

// -- update the x position of the decorations --
// used to center the rests
function deco_update(s, dx) {
    var	i, de,
	nd = a_de.length

	for (i = 0; i < nd; i++) {
		de = a_de[i]
		if (de.s == s)
			de.x += dx
	}
} // deco_update()

// -- adjust the left spacing of a symbol according to its decorations --
function deco_width(s, wlnt) {
    var	dd, i, w,
	wl = wlnt,
	wr = s.wr,
	a_dd = s.a_dd,
	nd = a_dd.length

	for (i = 0; i < nd; i++) {
		dd = a_dd[i]
		switch (dd.func) {
		case 1:			// slide
		case 2:			// arpeggio
			if (wl < 12)
				wl = 12
			break
//		case 3:
		case 5:
			switch (dd.glyph) {
			case "brth":
			case "caes":
			case "lphr":
			case "mphr":
			case "sphr":
			case "short":
			case "tick":
				if (s.wr < dd.wr)
					s.wr = dd.wr
				break
//			default:
//				w = dd.wl + 2
//				if (wl < w)
//					wl = w
//				break
			}
			// fall thru
		default:
			switch (dd.ty) {
			case '<':
				w = wlnt + dd.wl + dd.wr + 6
				if (wl < w)
					wl = w
				break
			case '>':
				w = wr + dd.wl + dd.wr + 6
				if (s.wr < w)
					s.wr = w
				break
			}
			break
		}
	}
	return wl
} // deco_width()

// compute the width of decorations in chord
function deco_wch(nt) {
    var	i, w, dd,
	wl = 0,
	n = nt.a_dd.length

	for (i = 0; i < n; i++) {
		dd = nt.a_dd[i]
		if (dd.ty == '<') {
			w = dd.wl + dd.wr + 4
			if (w > wl)
				wl = w
		}
	}
	return wl
} // deco_wch()

// -- draw the decorations --
// (the staves are defined)
Mus.prototype.draw_all_deco = function() {
	if (!a_de.length)
		return
    var	de, dd, s, note, f, st, x, y, y2, ym, uf, i, str, a,
	new_de = [],
	ymid = []

	st = nstaff
	y = staff_tb[st].y
	while (--st >= 0) {
		y2 = staff_tb[st].y
		ymid[st] = (y + 24 + y2) * .5
		y = y2
	}

	while (1) {
		de = a_de.shift()
		if (!de)
			break
		dd = de.dd
		if (!dd)
			continue		// deleted

		if (dd.dd_en)			// start of long decoration
			continue

		// handle the stem direction
		s = de.s
		f = dd.glyph
		i = f.indexOf('/')
		if (i > 0) {
			if (s.stem >= 0)
				f = f.slice(0, i)
			else
				f = f.slice(i + 1)
		}

		// no voice scale if staff decoration
		st = de.st
		if (f_staff[dd.func])
			set_sscale(st)
		else
			set_scale(s)

		if (!staff_tb[st].topbar)
			continue		// invisible staff
		x = de.x + (dd.dx || 0)
		y = de.y + staff_tb[st].y + (dd.dy || 0)

		// update the coordinates if head decoration
		if (de.m != undefined) {
			note = s.notes[de.m]
			if (note.shhd)
				x += note.shhd * stv_g.scale

		// center the dynamic marks between two staves
//fixme: KO when deco on other voice and same direction
		} else if (dd.func == 6
			&& ((de.pos & C.SL_ALI_MSK) == C.SL_CENTER
			 || ((de.pos & C.SL_ALI_MSK) == 0
			  && !s.fmt.dynalign))
			&& ((de.up && st > 0)
			 || (!de.up && st < nstaff))) {
			if (de.up)
				ym = ymid[--st]
			else
				ym = ymid[st++]
			ym -= dd.h * .5
			if ((de.up && y < ym)
			 || (!de.up && y > ym)) {
//				if (s.st > st) {
//					while (s.st != st)
//						s = s.ts_prev
//				} else if (s.st < st) {
//					while (s.st != st)
//						s = s.ts_next
//				}
				y2 = y_get(st, !de.up, de.x, de.val)
					+ staff_tb[st].y
				if (de.up)
					y2 -= dd.h
//fixme: y_set is not used later!
				if ((de.up && y2 > ym)
				 || (!de.up && y2 < ym)) {
					y = ym
//					y_set(st, de.up, de.x, de.val,
//						(de.up ? y + dd.h : y)
//							- staff_tb[st].y)
					if (stv_g.scale != 1)
						y += stv_g.dy / 2
				}
			}
		}

		// check if user JS decoration
		if (user.deco) {
			uf = user.deco[f]
			if (uf && typeof(uf) == "function") {
				uf.call(self, x, y, de)
				continue
			}
		}

		// check if user PS definition
		if (self.psdeco(x, y, de))
			continue

//		if (de.flags.grace) {
//			g_open(x, y, 0, .7, de.inv ? -.7 : 0)
//			x = y = 0
//		} else
		if (de.inv) {
			y = y + dd.h - dd.hd
			g_open(x, y, 0, 1, -1)
			x = y = 0
		}
		if (de.val != undefined) {
			if (dd.func != 2	// if not !arpeggio!
			 || stv_g.st < 0)	// or not staff scale
// || voice_tb[s.v].scale != 1)
				out_deco_val(x, y, f, de.val / stv_g.scale, de.defl)
			else
				out_deco_val(x, y, f, de.val, de.defl)
			if (de.cont)
				new_de.push(de.start)	// to be continued next line
		} else if (dd.str != undefined		// string
			&& !tgls[dd.glyph]
			&& !glyphs[dd.glyph]) {		// with a class
			out_deco_str(x, y, de)
		} else if (de.lden) {
			out_deco_long(x, y, de)
		} else {
			xygl(x, y, f)
		}
		if (stv_g.g)
			g_close()
		put_ref(s, 'deco')
	}

	// keep the long decorations which continue on the next line
	a_de = new_de
} // draw_all_deco()

// -- create the decorations and define the ones near the notes --
// (the staves are not yet defined)
// (delayed output)
// this function must be called first as it builds the deco element table
function draw_deco_near() {
    var	s, g

	// update starting old decorations
	function ldeco_update(s) {
	    var	i, de,
//		x = s.ts_prev.x + s.ts_prev.wr
		x = s.x - s.wl,
		nd = a_de.length

		for (i = 0; i < nd; i++) {
			de = a_de[i]
			de.ix = i
			de.s.x = de.x = x
			de.defl.nost = 1 //true
		}
	}

	// -- create the deco elements, and treat the near ones --
	function create_deco(s) {
	    var	dd, k, pos, de, x, y, up,
		nd = s.a_dd.length

		if (s.y == undefined)
			s.y = 0			// (no y in measure bars)

//fixme:pb with decorations above the staff
		for (k = 0; k < nd; k++) {
			dd = s.a_dd[k]

			// adjust the position
			x = s.x
			y = s.y
			switch (dd.func) {
			default:
				if (dd.func >= 10)
					continue
				pos = 0
				break
			case 3:				// d_upstaff
			case 4:
			case 5:				// after slurs
				pos = s.pos.orn
				break
			case 6:				// dynamic
				pos = s.pos.dyn
				break
			}

			switch (dd.ty) {		// explicit position
			case '^':
				pos = (pos & ~0x07) | C.SL_ABOVE
				break
			case '_':
				pos = (pos & ~0x07) | C.SL_BELOW
				break
			case '<':
			case '>':
				pos = (pos & 0x07) | C.SL_CLOSE
				if (dd.ty == '<') {
					x -= dd.wr + 8
					if (s.notes[0].acc)
						x -= 8
				} else {
					x += dd.wl + 8
				}
				y = 3 * s.notes[0].pit
						- (dd.h - dd.hd) / 2
				break
			case '@':
				x += dd.x
				y += dd.y
				break
			}

			if ((pos & 0x07) == C.SL_HIDDEN)
				continue

			de = {
				s: s,
				dd: dd,
				st: s.st,
				ix: a_de.length,
				defl: {},
				x: x,
				y: y,
				pos: pos
//				dy: 0
			}

			up = 0 //false
			if (dd.ty == '^') {
				up = 1 //true
			} else if (dd.ty == '_') {
				;
			} else {
				switch (dd.func) {
				case 0:
					if (s.multi)
						up = s.multi > 0
					else
						up = s.stem < 0
					break
				case 3:
				case 5:
					up = up3(s, pos)
					break
				case 6:
				case 7:
					up = up6(s, pos)
					break
				}
			}
			de.up = up

			if (dd.name.indexOf("inverted") >= 0)
				de.inv = 1 //true

			if (s.bar_type && !dd.ty)
				de.x -= s.wl / 2 - 2
			a_de.push(de)
			if (dd.dd_en) {
				de.ldst = 1 //true
			} else if (dd.dd_st) {
//fixme: pb with "()"
				de.lden =
					de.defl.nost = 1 //true
			}

			if (f_near[dd.func])
				f_near[dd.func](de)
		}
	} // create_deco()

	// create the decorations of note heads
	function create_dh(s, m) {
	    var	de, k, dd,
		note = s.notes[m],
		nd = note.a_dd.length,
		x = s.x

		for (k = 0; k < nd; k++) {
			dd = note.a_dd[k]
//fixme: check if hidden?
			de = {
				s: s,
				dd: dd,
				st: s.st,
				m: m,
				ix: 0,
				defl: {},
				x: x,
				y: 3 * note.pit - (dd.h - dd.hd) / 2
//				dy: 0
			}

			if (dd.ty) {		// if explicit position
				if (dd.ty == '@') {
					de.x += dd.x
					de.y += dd.y
				} else {
					de.y -= (dd.h - dd.hd) / 2	// center
					if (dd.ty == '<') {
						de.x -= dd.wr + 8
						if (s.notes[m].acc)
							x -= 8
					} else if (dd.ty == '>') {
						de.x += dd.wl + 8
					}
				}
			}

			a_de.push(de)
			if (dd.dd_en) {
				de.ldst = 1 //true
			} else if (dd.dd_st) {
				de.lden =
					de.defl.nost = 1 //true
			}
		}
	} // create_dh()

	// create all decorations of a note (chord and heads)
	function create_all(s) {
		if (s.invis && s.play)	// play sequence: no decoration
			return
		if (s.a_dd)
			create_deco(s)
		if (s.notes) {
			for (var m = 0; m < s.notes.length; m++) {
				if (s.notes[m].a_dd)
					create_dh(s, m)
			}
		}
	} // create_all()

	// link the long decorations
	function ll_deco() {
	    var	i, j, de, de2, de3, dd, dd2, v, s, st,
		n_de = a_de.length

		// add ending decorations
		for (i = 0; i < n_de; i++) {
			de = a_de[i]
			if (!de.ldst)	// not the start of long decoration
				continue
			dd = de.dd
			dd2 = dd.dd_en
			s = de.s
			v = s.v			// search later in the voice
			for (j = i + 1; j < n_de; j++) {
				de2 = a_de[j]
				if (!de2.start
				 && de2.dd == dd2 && de2.s.v == v)
					break
			}
			if (j == n_de) {	// no end, search in the staff
				st = s.st
				for (j = i + 1; j < n_de; j++) {
					de2 = a_de[j]
					if (!de2.start
					 && de2.dd == dd2 && de2.s.st == st)
						break
				}
			}
			if (j == n_de) {	// no end, insert one
				de2 = {
					s: s,
					st: de.st,
					dd: dd2,
					ix: a_de.length - 1,
					x: realwidth - 6,
					y: s.y,
					cont: 1,	// keep for next line
					lden: 1, //true
					defl: {
						noen: 1 //true
					}
				}
				if (de2.x < s.x + 10)
					de2.x = s.x + 10
				if (de.m != undefined)
					de2.m = de.m
				a_de.push(de2)
			}
			de2.start = de
			de2.defl.nost = de.defl.nost

			// handle same numbered decoration at a same time
			j = i
			while (--j >= 0) {
				de3 = a_de[j]
				if (!de3.start)
					continue
				if (de3.s.time < s.time)
					break
				if (de3.dd.name == de2.dd.name) {
					de2.prev = de3
					break
				}
			}
		}

		// add starting decorations
		for (i = 0; i < n_de; i++) {
			de2 = a_de[i]
			if (!de2.lden	// not the end of long decoration
			 || de2.start)	// start already found
				continue
			s = de2.s
			de = {
				s: prev_scut(s),
				st: de2.st,
				dd: de2.dd.dd_st,
				ix: a_de.length - 1,
//				x: s.x - s.wl - 4,
				y: s.y,
				ldst: 1
			}
			de.x = de.s.x + de.s.wr
			if (de2.m != undefined)
				de.m = de2.m
			a_de.push(de)
			de2.start = de
		}
	} // ll_deco

	// update the long decorations started in the previous line
	for (s = tsfirst; s; s = s.ts_next) {
		switch (s.typ) {
		case C.clef:
		case C.key:
		case C.meter:
			continue
		}
		break
	}
	if (a_de.length)
		ldeco_update(s)

	for ( ; s; s = s.ts_next) {
		switch (s.typ) {
		case C.bar:
		case C.mrest:
		case C.note:
		case C.rest:
		case C.space:
			break
		case C.grace:
			for (g = s.grasym; g; g = g.next)
				create_all(g)
			break
		default:
			continue
		}
		create_all(s)
	}
	ll_deco()			// link the long decorations
} // draw_deco_near()

// -- define the decorations tied to a note --
// (the staves are not yet defined)
// (delayed output)
function draw_deco_note() {
    var	i, de, dd, f,
	nd = a_de.length

	for (i = 0; i < nd; i++) {
		de = a_de[i]
		dd = de.dd
		f = dd.func
		if (f_note[f]
		 && de.m == undefined)
			f_note[f](de)
	}
} // draw_deco_note()

// -- define the decorations tied to the staff --
// (the staves are not yet defined)
// (unscaled delayed output)
function draw_deco_staff() {
    var	s, p_voice, y, i, v, de, dd, w,
	minmax = new Array(nstaff + 1),
	nd = a_de.length

	// create the decorations tied to the staves
	for (i = 0; i <= nstaff; i++)
		minmax[i] = {
			ymin: 0,
			ymax: 0
		}
	for (i = 0; i < nd; i++) {
		de = a_de[i]
		dd = de.dd
		if (!dd)		// if error
			continue
		if (!f_staff[dd.func]	// if not tied to the staff
		 || de.m != undefined	// or head decoration
		 || dd.ty == '<' || dd.ty == '>' || dd.ty == '@')
			continue

		f_staff[dd.func](de)
		if (dd.func != 6
		 || dd.dd_en)		// if start
			continue

		if ((de.pos & C.SL_ALI_MSK) == C.SL_ALIGN
		 || ((de.pos & C.SL_ALI_MSK) == 0
		  && de.s.fmt.dynalign > 0)) {	// if align
			if (de.up) {
				if (de.y > minmax[de.st].ymax)
					minmax[de.st].ymax = de.y
			} else {
				if (de.y < minmax[de.st].ymin)
					minmax[de.st].ymin = de.y
			}
		}
	}

	// set the same vertical offset of the dynamic marks
	for (i = 0; i < nd; i++) {
		de = a_de[i]
		dd = de.dd
		if (!dd)				// if error
			continue

		// if @x,y offsets, update the top and bottom of the staff
		if (dd.ty == '@') {
		    var	y2

			y = de.y
			if (y > 0) {
				y2 = y + dd.h + 2
				if (y2 > staff_tb[de.st].ann_top)
					staff_tb[de.st].ann_top = y2
			} else {
				y2 = y - dd.hd - 2
				if (y2 < staff_tb[de.st].ann_bot)
					staff_tb[de.st].ann_bot = y2
			}
			continue
		}
		if (dd.func != 6
		 || dd.ty == '<' || dd.ty == '>' || dd.ty == '@'
		 || dd.dd_en)				// if start
			continue

		w = de.val || (dd.wl + dd.wr)
		if ((de.pos & C.SL_ALI_MSK) == C.SL_ALIGN
		 || ((de.pos & C.SL_ALI_MSK) == 0
		  && de.s.fmt.dynalign > 0)) {		// if align
			de.y =
				y = minmax[de.st][de.up ? "ymax" : "ymin"]
		} else {
			y = de.y
		}
		y_set(de.st, de.up, de.x, w,
				de.up ? (y + dd.h) : (y - dd.hd))
	}

	// second pass for pedal (under the staff)
	for (i = 0; i < nd; i++) {
		de = a_de[i]
		dd = de.dd
		if (!dd)			// if error
			continue
		if (dd.dd_en			// if start
		 || dd.name.slice(0, 3) != "ped")
			continue
		w = de.val || 10
		de.y = y_get(de.st, 0, de.x, w)
			- (dd.dd_st && cfmt.pedline ? 10 : dd.h)
		y_set(de.st, 0, de.x, w, de.y)	// (no descent)
	}
} // draw_deco_staff()

// -- draw the measure bar numbers --
// (scaled delayed output)
function draw_measnb() {
    var	s, st, bar_num, x, y, w, any_nb, font_size, w0, s2,
	sy = cur_sy

	// search the top staff
	for (st = 0; st <= nstaff; st++) {
		if (sy.st_print[st])
			break
	}
	if (st > nstaff)
		return				// no visible staff
	set_dscale(st)

	// leave the measure numbers as unscaled
	if (staff_tb[st].staffscale != 1) {
		font_size = get_font("measure").size
		param_set_font("measurefont", "* " +
			(font_size / staff_tb[st].staffscale).toString())
	}
	set_font("measure")
	w0 = cwidf('0')			// (greatest) width of a number

	s = tsfirst				// clef
	bar_num = gene.nbar
	if (bar_num > 1) {
		if (!cfmt.measurenb) {
			any_nb = 1 //true
			y = y_get(st, 1, 0, 20)
			if (y < staff_tb[st].topbar + 14)
				y = staff_tb[st].topbar + 14
			xy_str(0, y - gene.curfont.size * .22, bar_num.toString())
			y_set(st, 1, 0, 20, y + gene.curfont.size + 2)
		} else if (bar_num % cfmt.measurenb == 0) {
			for ( ; ; s = s.ts_next) {
				switch (s.typ) {
				case C.clef:
				case C.key:
				case C.meter:
				case C.stbrk:
					continue
				}
				break
			}

			// don't display the number twice
			if (!s.bar_type || !s.bar_num) {
				any_nb = 1 //true
				w = w0
				if (bar_num >= 10)
					w *= bar_num >= 100 ? 3 : 2
				if (gene.curfont.pad)
					w += gene.curfont.pad * 2
				x = (s.prev
					? s.prev.x + s.prev.wr / 2
					: s.x - s.wl) - w
				y = y_get(st, 1, x, w) + 5
				if (y < staff_tb[st].topbar + 6)
					y = staff_tb[st].topbar + 6
				y += gene.curfont.pad
				xy_str(x, y - gene.curfont.size * .22,
					bar_num.toString())
				y += gene.curfont.size + gene.curfont.pad
				y_set(st, 1, x, w, y)
//				s.ymx = y
			}
		}
	}

	for ( ; s; s = s.ts_next) {
		switch (s.typ) {
		case C.staves:
			sy = s.sy
			for (st = 0; st < nstaff; st++) {
				if (sy.st_print[st])
					break
			}
			set_dscale(st)
			continue
		default:
			continue
		case C.bar:
			if (!s.bar_num || s.bar_num <= 1)
				continue
			break
		}

		s2 = s.next

		bar_num = s.bar_num
		if (!cfmt.measurenb
		 || (bar_num % cfmt.measurenb) != 0
		 || !s2
		 || s.bar_mrep)
			continue
		if (!any_nb)
			any_nb = 1 //true
		w = w0
		if (bar_num >= 10)
			w *= bar_num >= 100 ? 3 : 2
		if (gene.curfont.pad)
			w += gene.curfont.pad * 2
		x = s.x
		y = y_get(st, 1, x, w)
		if (y < staff_tb[st].topbar + 6)
			y = staff_tb[st].topbar + 6
		if (s2.typ == C.note) {
			if (s2.stem > 0) {
				if (y < s2.ys - gene.curfont.size)
					y = s2.ys - gene.curfont.size
			} else {
				if (y < s2.y)
					y = s2.y
			}
		}
		y += 2 + gene.curfont.pad
		xy_str(x, y - gene.curfont.size * .22, bar_num.toString())
		y += gene.curfont.size + gene.curfont.pad
		y_set(st, 1, x, w, y)
//		s.ymx = y
	}
	gene.nbar = bar_num

	if (font_size)
		param_set_font("measurefont", "* " + font_size.toString())
} // draw_measnb()

// -- draw the parts and the tempo information --
// (the staves are being defined)
function draw_partempo() {
    var	s, s2, s_part, s_tempo, h, w, y, st,
	sy = cur_sy

	// search the top staff
	for (st = 0; st <= nstaff; st++) {
		if (sy.st_print[st])
			break
	}
	if (st > nstaff)
		return				// no visible staff
	set_dscale(st, 1)			// no scale

	// get the minimal y offset
    var	ymin = staff_tb[st].topbar + 2,
	dosh = 0,
	shift = 1,
	x = -100,			// (must be negative for %%soloffs)
	yn = 0				// y min when x < 0

	// output the parts
	for (s = tsfirst; s; s = s.ts_next) {
		if (s.typ != C.part || s.invis)
			continue
		if (!s_part) {
			s_part = s
			set_font("parts")
			h = gene.curfont.size + 2 + gene.curfont.pad * 2
		}
		if (s.x == undefined) {
			s2 = s
			while (!s2.seqst)
				s2 = s2.ts_prev
			s.x = s2.x
		}
		s.x -= 10			// shift a bit
		w = strwh(s.text)[0]
		y = y_get(st, 1, s.x, w + 3)
		if (ymin < y)
			ymin = y
	}
	if (s_part) {
		set_sscale(-1)
		ymin += 2
		ymin *= staff_tb[st].staffscale
		for (s = s_part; s; s = s.ts_next) {
			if (s.typ != C.part || s.invis)
				continue
			w = strwh(s.text)[0]
			xy_str(s.x,
				ymin + gene.curfont.pad + gene.curfont.size * .22,
				s.text)
			if (user.annotate) {
				s.st = s.wl = 0
				s.wr = w
				s.ymn = ymin
				s.ymx = s.ymn + h
				anno_a.push(s)
			}
			y_set(st, 1, s.x, w + 3, ymin + h)
			if (s.x < 0)
				yn = ymin + h
		}
	}

	// output the tempos
	ymin = staff_tb[st].topbar + 6
	for (s = tsfirst; s; s = s.ts_next) {
		if (s.typ != C.tempo || s.invis)
			continue
		if (!s_tempo)
			s_tempo = s
		w = s.tempo_wh[0]
//		if (!s.time && s.x > 40)	// at start of tune and no %%soloffs,
//			s.x = 40	// shift the tempo over the key signature
		y = y_get(st, 1, s.x - 16, w)
		if (s.x - 16 < 0)
			y = yn
		if (y > ymin)
			ymin = y
		if (x >= s.x - 16 && !(dosh & (shift >> 1)))
			dosh |= shift
		shift <<= 1
		x = s.x - 16 + w
	}
	if (s_tempo) {
		set_sscale(-1)
		set_font("tempo")
		h = gene.curfont.size
		ymin += 2
		ymin *= staff_tb[st].staffscale

		// draw the tempo indications
		for (s = s_tempo; s; s = s.ts_next) {
			if (s.typ != C.tempo
			 || s.invis)		// (displayed by %%titleformat)
				continue
			w = s.tempo_wh[0]
			y = ymin
			if (dosh & 1)
				y += h
			writempo(s, s.x - 16, y)
			if (user.annotate) {
				s.wl = 16
				s.wr = w - 16	//was 30
				s.ymn = y
				s.ymx = s.ymn + 14
				anno_a.push(s)
			}
			y_set(st, 1, s.x - 16, w, y + h + 2)
			dosh >>= 1
		}
	}
} // draw_partempo()
