// txtmus - format.js - formatting functions
//
// Copyright (C) 2014-2024 Jean-Francois Moine
//
// This file is part of txtmus-core.
//
// txtmus-core is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus-core is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus-core.  If not, see <http://www.gnu.org/licenses/>.

    var	font_scale_tb = {
		serif: 1,
		serifBold: 1,
		'sans-serif': 1,
		'sans-serifBold': 1,
		Palatino: 1.1,
		monospace: 1
	},
	txt_s = "ui-serif,serif",		// text font-families
	txt_ss = "ui-sans-serif,sans-serif",
	fmt_lock = {}

    var	cfmt = {
	"abc-version": "1",		// default: old version
	annotationfont: {name: txt_s, size: 12},
	aligncomposer: 1,
	beamslope: .4,			// max slope of a beam
//	botmargin: .7 * IN,		// != 1.8 * CM,
	bardef: {
		"[":	"",		// invisible
		"[]":	"",
		"|:":	"[|:",
		"|::":	"[|::",
		"|:::":	"[|:::",
		":|":	":|]",
		"::|":	"::|]",
		":::|":	":::|]",
		"::":	":][:"
	},
	breaklimit: .7,
	breakoneoln: 1, //true,
	cancelkey: 1, //true,
	composerfont: { name: txt_s, style: "italic", size: 14 },
	composerspace: 6,
//	contbarnb: false,
	decoerr: 1, //true,
	dynalign: 1, //true,
	fullsvg: '',
	gchordfont: { name: txt_ss, size: 12 },
	gracespace: new Float32Array([6, 8, 11]),	// left, inside, right
	graceslurs: 1, //true,
	historyfont: { name: txt_s, size: 16 },
	hyphencont: 1, //true,
	indent: 0,
	infofont: {name: txt_s, style: "italic", size: 14 },
	infoname: 'R "Rhythm: "\n\
B "Book: "\n\
S "Source: "\n\
D "Discography: "\n\
N "Notes: "\n\
Z "Transcription: "\n\
H "History: "',
	infospace: 0,
	keywarn: 1, //true,
	leftmargin: 1.4 * CM,
	lineskipfac: 1.1,
	linewarn: 1, //true,
	maxshrink: .65,		// nice scores
	maxstaffsep: 2000,
	maxsysstaffsep: 2000,
	measrepnb: 1,
	measurefont: {name: txt_s, style: "italic", size: 10},
	measurenb: -1,
	musicfont: {name: "music", src: musicfont, size: 24},
	musicspace: 6,
//	notespacingfactor: "1.33, 40",
	partsfont: {name: txt_s, size: 15},
	parskipfac: .4,
	partsspace: 8,
//	pageheight: 29.7 * CM,
	pagewidth: 21 * CM,
	"propagate-accidentals": "o",		// octave
	rightmargin: 1.4 * CM,
	rbmax: 4,
	rbmin: 2,
	repeatfont: {name: txt_s, size: 9},
	scale: 1,
	slurheight: 1.0,
	spc_tb: 	// spacing table (see "notespacingfactor" and set_space())
		new Float32Array([	// default = "1.33, 40"
			9.6, 12.8, 17.0, 22.6, 30.0,
			40,
			53.2, 70.8, 94.1, 125.1 ]),
	staffsep: 46,
	stemheight: 21,			// one octave
	stretchlast: .25,
	stretchstaff: 1, //true,
	subtitlefont: {name: txt_s, size: 16},
	subtitlespace: 3,
	sysstaffsep: 34,
	systnames: -1,
	tempofont: {name: txt_s, weight: 700, size: 12},
	textfont: {name: txt_s, size: 16},
//	textoption: undefined,
	textspace: 14,
	tieheight: 1.0,
	titlefont: {name: txt_s, size: 20},
//	titleleft: false,
	titlespace: 6,
	titletrim: 1, //true,
//	transp: 0,			// global transpose
//	topmargin: .7 * IN,
	topspace: 22,
	tuplets: [0, 0, 0, 0],
	tupletfont: {name: txt_s, style: "italic", size: 10},
	vocalfont: {name: txt_s, weight: 700, size: 13},
	vocalspace: 10,
	voicefont: {name: txt_s, weight: 700, size: 13},
//	voicescale: 1,
	writefields: "CMOPQsTWw",
	wordsfont: {name: txt_s, size: 16},
	wordsspace: 5,
	"writeout-accidentals": "n"
} // cfmt

// parameters that are used in the symbols
    var	sfmt = "\
bardef,barsperstaff,beamslope,breaklimit,bstemdown,\
cancelkey,\
dynalign,\
flatbeams,\
gracespace,\
hyphencont,\
keywarn,\
maxshrink,maxstaffsep,measrepnb,\
rbmax,rbmin,\
shiftunison,slurheight,squarebreve,staffsep,stemheight,stretchlast,stretchstaff,\
systnames, tieheight,timewarn,trimsvg,\
vocalspace"
	// sfmt

function get_bool(parm) {
	return !parm || !/^(0|n|f)/i.test(parm) // accept void as true !
}

// %%font <font> [<encoding>] [<scale>]
function get_font_scale(parm) {
    var	i, font,
	a = info_split(parm)	// a[0] = font name

	if (a.length <= 1)
		return
    var	sc = +a[a.length - 1]

	if (isNaN(sc) || sc <= 0.5) {
		syntax(1, "Bad scale value in %%font")
		return
	}
	font_scale_tb[a[0]] = sc
} // get_font_scale()

// set the width factor of a font
function set_font_fac(font) {
    var sc = font_scale_tb[font.fname || font.name]

	if (!sc)
		sc = 1.1
	font.swfac = font.size * sc
} // set_font_fac()

// get the class index of the font (.fx)
function get_fx(sty) {
    var	fx = tm.styfx[sty]

	if (fx == undefined) {
		fx = tm.frfx++
		tm.styfx[sty] = fx
	}
	return fx
} // get_ix()

// %%xxxfont fontname|* [encoding] [size|*]
function param_set_font(xxxfont, p) {
    var	font, n, a, ft2, k

	// "setfont-<n>" goes to "u<n>font"
	if (xxxfont[xxxfont.length - 2] == '-') {
		n = xxxfont[xxxfont.length - 1]
		if (n < '1' || n > '9')
			return
		xxxfont = "u" + n + "font"
	}

	// fill the values
	font = {}
	a = p.match(/\s+(no)?box(\s|$)/)
	if (a) {				// if box
		if (a[1]) {
			font.box = 0 //false	// nobox
			font.pad = 0
		} else {
			font.box = 1 //true
			font.pad = 2.5
		}
		p = p.replace(a[0], a[2])
	}
	a = p.match(/\s+padding=([\d.]+)(\s|$)/)
	if (a) {				// if padding
		font.pad = a[1] ? +a[1] : 0
		p = p.replace(a[0], a[2])
	}

	a = p.match(/\s+class=(.*?)(\s|$)/)
	if (a) {
		font.class = a[1]
		p = p.replace(a[0], a[2])
	}
	a = p.match(/\s+wadj=(.*?)(\s|$)/)
	if (a) {
	    if (typeof document == "undefined")	// useless if in browser
		switch (a[1]) {
		case 'none':
			font.wadj = ''
			break
		case 'space':
			font.wadj = 'spacing'
			break
		case 'glyph':
			font.wadj = 'spacingAndGlyphs'
			break
		default:
			syntax(1, errs.bad_val, "%%" + xxxfont)
			break
		}
		p = p.replace(a[0], a[2])
	}

	// the font size is the last item
	a = p.match(/\s+([0-9.]+|\*)$/)
	if (a) {
		if (a[1] != "*")
			font.size = +a[1]
		p = p.replace(a[0], '')
	}

	// accept local(..) and url(...) as the font source
	if ((p[0] == 'u' && p.slice(0, 4) == "url(")
	 || (p[0] == 'l' && p.slice(0, 6) == "local(")) {
		n = p.indexOf(')', 1)
		if (n < 0) {
			syntax(1, "No end of url in font family")
			return
		}

		font.src = p.slice(0, n + 1)
		font.fx = get_fx(p)
		font.name = 'ft' + font.fx
		p = p.replace(font.src, '')
	}

	// extract the font attributes
	a = p.match(/[- ]?[nN]ormal/)
	if (a) {
		font.normal = 1 //true
		p = p.replace(a[0], '')
	}

	// font weight
	a = p.match(tm.ft_re)
	if (a) {
		font.weight = tm.ft_w[a[0].replace(/[ -]/, '')
				.toLowerCase()]
		p = p.replace(a[0], '')
	}

	a = p.match(/[- ]?[iI]talic/)
	if (a) {
		font.style = "italic"
		p = p.replace(a[0], '')
	}
	a = p.match(/[- ]?[oO]blique/)
	if (a) {
		font.style = "oblique"
		p = p.replace(a[0], '')
	}

	if (!font.src) {			// if no url(...)
		if (p[0] == '"') {
			n = p.indexOf('"', 1)
			if (n < 0) {
				syntax(1, "No end of string in font family")
				return
			}
			p = p.slice(1, n)
		}

		// here is the font family
		p = p.trim()

		switch (p) {
		case "":
		case "*": p = ""; break
		case "Times-Roman":
		case "Times":	p = txt_s; break
		case "Helvetica": p = txt_ss; break
		case "Courier": p = "ui-monospace,monospace"; break
		case "music": p = cfmt.musicfont.name; break
		default:
//hack: the font "Figurato" is used for figured bass
			if (p.indexOf("Fig") > 0)
				font.figb = 1 //true
			break
		}
	}
	if (p && !font.name)
		font.name = p

	if (font.size)
		set_font_fac(font)

	// keep the previous attributes if no font name nor size
	if (!font.name || !font.size) {
		ft2 = cfmt[xxxfont]
		for (k in ft2) {
			if (!ft2.hasOwnProperty(k)
			 || font[k] != undefined)
				continue
			switch (k) {
			case "fx":
			case "src":
				break
			case "style":
			case "weight":
				if (font.normal)
					break
				// fall thru
			default:
				font[k] = ft2[k]
				break
			}
		}
		if (!font.swfac)
			set_font_fac(font)
	}
	if (font.pad == undefined)
		font.pad = 0
	font.fname = font.name
	if (font.weight >= 700)
		font.fname += 'Bold'

	cfmt[xxxfont] = font
} // param_set_font()

// get a length with a unit - return the number of pixels
function get_unit(parm) {
    var	v = parm.toLowerCase().match(/(-?[\d.]+)(.*)/)
	if (!v)
		return NaN

	v[1] = +v[1]
	switch (v[2]) {
	case "cm":
		return v[1] * CM
	case "in":
		return v[1] * IN
	case "pt":		// paper point in 1/72 inch
		return v[1] / .75
	case "px":		// screen pixel in 1/96 inch
	case "":
		return v[1]
	}
	return NaN
}

// set the infoname
function set_infoname(parm) {
//fixme: check syntax: '<letter> ["string"]'
    var	tmp = cfmt.infoname.split("\n"),
	letter = parm[0]

	for (var i = 0; i < tmp.length; i++) {
	    var	infoname = tmp[i]
		if (infoname[0] != letter)
			continue
		if (parm.length == 1)
			tmp.splice(i, 1)
		else
			tmp[i] = parm
		cfmt.infoname = tmp.join('\n')
		return
	}
	cfmt.infoname += "\n" + parm
}

// get the text option
function get_textopt(v) {
    var	i = v.indexOf(' ')
	if (i > 0)
		v = v.slice(0, i)
	return tm.textopt[v]
} // get_textopt()

// -- set the position of elements in a voice --
function set_pos(k, v) {		// keyword, value
	k = k.slice(0, 3)
	if (k == "ste")
		k = "stm"
	set_v_param("pos", '"' + k + ' ' + v + '"')
} // set_pos()

// set/unset the fields to write
function set_writefields(parm) {
    var	c, i,
	a = parm.split(/\s+/)

	if (get_bool(a[1])) {
		for (i = 0; i < a[0].length; i++) {	// set
			c = a[0][i]
			if (cfmt.writefields.indexOf(c) < 0)
				cfmt.writefields += c
		}
	} else {
		for (i = 0; i < a[0].length; i++) {	// unset
			c = a[0][i]
			if (cfmt.writefields.indexOf(c) >= 0)
				cfmt.writefields = cfmt.writefields.replace(c, '')
		}
	}
} // set_writefields()

// set a voice specific parameter
function set_v_param(k, v) {
	k = [k, '=', v]
	if (parse.state < 3)
		memo_kv_parm(curvoice ? curvoice.id : '*', k)
	else if (curvoice)
		set_kv_parm(k)
	else
		memo_kv_parm('*', k)
} // set_v_param()

function set_page() {
	if (!img.chg)
		return
	img.chg = 0 //false
	img.lm = cfmt.leftmargin
	img.rm = cfmt.rightmargin
	img.width = cfmt.pagewidth

	// must have 100pt at least as the staff width
	if (img.width - img.lm - img.rm < 100) {
		error(0, undefined, "Bad staff width")
		img.width = img.lm + img.rm + 150
	}
	set_posx()
} // set_page()

// set a format parameter
// (possible hook)
Mus.prototype.set_format = function(cmd, parm) {
    var	f, f2, v, i

//fixme: should check the type and limits of the parameter values
	if (/.+font(-[\d])?$/.test(cmd)) {
		if (cmd == "soundfont")
			cfmt.soundfont = parm
		else
			param_set_font(cmd, parm)
		return
	}

	// duplicate the global parameters if already used by symbols
	if (sfmt.indexOf(cmd) >= 0 && parse.ufmt)
		cfmt = Object.create(cfmt)

	switch (cmd) {
	case "aligncomposer":
	case "barsperstaff":
	case "infoline":
	case "measurenb":
	case "rbmax":
	case "rbmin":
	case "measrepnb":
	case "shiftunison":
	case "systnames":
		v = +parm
		if (isNaN(v)) {
			syntax(1, "Bad integer value")
			break
		}
		cfmt[cmd] = v
		break
	case "abc-version":
	case "bgcolor":
	case "fgcolor":
	case "propagate-accidentals":
	case "titleformat":
	case "writeout-accidentals":
		cfmt[cmd] = parm
		break
	case "beamslope":
	case "breaklimit":			// float values
	case "lineskipfac":
	case "maxshrink":
	case "pagescale":
	case "parskipfac":
	case "scale":
	case "slurheight":
	case "stemheight":
	case "tieheight":
		f = +parm
		if (isNaN(f) || !parm || f < 0) {
			syntax(1, errs.bad_val, '%%' + cmd)
			break
		}
		switch (cmd) {
		case "scale":			// old scale
			f /= .75
		case "pagescale":
			if (f < .1)
				f = .1		// smallest scale
			cmd = "scale"
			img.chg = 1 //true
			break
		}
		cfmt[cmd] = f
		break
	case "annotationbox":
	case "gchordbox":
	case "measurebox":
	case "partsbox":
		param_set_font(cmd.replace("box", "font"),	// font
			"* * " + (get_bool(parm) ? "box" : "nobox"))
		break
	case "altchord":
	case "bstemdown":
	case "breakoneoln":
	case "cancelkey":
	case "checkbars":
	case "contbarnb":
	case "custos":
	case "decoerr":
	case "flatbeams":
	case "graceslurs":
	case "graceword":
	case "hyphencont":
	case "keywarn":
	case "linewarn":
	case "quiet":
	case "squarebreve":
	case "splittune":
	case "straightflags":
	case "stretchstaff":
	case "timewarn":
	case "titlecaps":
	case "titleleft":
	case "trimsvg":
		cfmt[cmd] = get_bool(parm)
		break
	case "dblrepbar":
		parm = ":: " + parm
		// fall thru
	case "bardef":			// %%bardef oldbar newbar
		v = parm.split(/\s+/)
		if (v.length != 2) {
			syntax(1, errs.bad_val, "%%bardef")
		} else {
			if (parse.ufmt)
				cfmt.bardef = Object.create(cfmt.bardef) // new object
			cfmt.bardef[v[0]] = v[1]
		}
		break
	case "chordalias":
		v = parm.split(/\s+/)
		if (!v.length)
			syntax(1, errs.bad_val, "%%chordalias")
		else
			tm.ch_alias[v[0]] = v[1] || ""
		break
	case "composerspace":
	case "indent":
	case "infospace":
	case "maxstaffsep":
	case "maxsysstaffsep":
	case "musicspace":
	case "partsspace":
	case "staffsep":
	case "subtitlespace":
	case "sysstaffsep":
	case "textspace":
	case "titlespace":
	case "topspace":
	case "vocalspace":
	case "wordsspace":
		f = get_unit(parm)	// normally, unit in points - 72 DPI accepted
		if (isNaN(f))
			syntax(1, errs.bad_val, '%%' + cmd)
		else
			cfmt[cmd] = f
		break
	case "page-format":
		user.page_format = get_bool(parm)
		break
//	case "botmargin":
	case "leftmargin":
//	case "pageheight":
	case "pagewidth":
	case "rightmargin":
//	case "topmargin":
		f = get_unit(parm)	// normally unit in cm or in - 96 DPI
		if (isNaN(f)) {
			syntax(1, errs.bad_val, '%%' + cmd)
			break
		}
		cfmt[cmd] = f
		img.chg = 1 //true
		break
	case "concert-score":
		cfmt.sound = "concert"
		break
	case "writefields":
		set_writefields(parm)
		break
	case "volume":
		cmd = "dynamic"
		// fall thru
	case "dynamic":
	case "gchord":
	case "gstemdir":
	case "ornament":
	case "stemdir":
	case "vocal":
		set_pos(cmd, parm)
		break
	case "font":
		get_font_scale(parm)
		break
	case "fullsvg":
		if (parse.state != 0) {
			syntax(1, errs.not_in_tune, "%%fullsvg")
			break
		}
//fixme: should check only alpha, num and '_' characters
		cfmt[cmd] = parm
		break
	case "gracespace":
		v = parm.split(/\s+/)
		for (i = 0; i < 3; i++)
			if (isNaN(+v[i])) {
				syntax(1, errs.bad_val, "%%gracespace")
				break
			}
		if (parse.ufmt)
			cfmt[cmd] = new Float32Array(3)
		for (i = 0; i < 3; i++)
			cfmt[cmd][i] = +v[i]
		break
	case "tuplets":
		v = parm.split(/\s+/)
		f = v[3]
		if (f)				// if 'where'
			f = tm.posval[f]	// translate the keyword
		v[3] = f || 0
		v = new Int8Array(v)
		if (curvoice)
			curvoice.tup = v
		else
			cfmt[cmd] = v
		break
	case "infoname":
		set_infoname(parm)
		break
	case "notespacingfactor":
		v = parm.match(/([.\d]+)[,\s]*(\d+)?/)
		if (v) {
			f = +v[1]
			if (isNaN(f) || f < 1 || f > 2) {
				f = 0
			} else if (v[2]) {
				f2 = +v[2]
				if (isNaN(f))
					f = 0
			} else {
				f2 = cfmt.spc_tb[5]
			}
		}
		if (!f) {
			syntax(1, errs.bad_val, "%%" + cmd)
			break
		}
		cfmt[cmd] = parm		// (for dump)

		// in the table 'spc_tb',
		// the width of notes is indexed by log2(note_length)
		cfmt.spc_tb = new Float32Array(10)
		i = 5				// index of crotchet
		do {
			cfmt.spc_tb[i] = f2
			f2 /= f
		} while (--i >= 0)
		i = 5
		f2 = cfmt.spc_tb[i]
		for ( ; ++i < cfmt.spc_tb.length; ) {
			f2 *= f
			cfmt.spc_tb[i] = f2
		}
		break
	case "pos":
		cmd = parm.match(/(\w*)\s+(.*)/)
		if (!cmd || !cmd[2]) {
			syntax(1, "Error in %%pos")
			break
		}
		if (cmd[1].slice(0, 3) == 'tup'		// special case for %%pos tuplet
		 && curvoice) {				// inside tune
			if (!curvoice.tup)
				curvoice.tup = cfmt.tuplets
			else
				curvoice.tup = Object.create(curvoice.tup)
			v = tm.posval[cmd[2]]
			switch (v) {
			case C.SL_ABOVE:
				curvoice.tup[3] = 1
				break
			case C.SL_BELOW:
				curvoice.tup[3] = 2
				break
			case C.SL_HIDDEN:
				curvoice.tup[2] = 1
				break
			}
			break
		}
		if (cmd[1].slice(0, 3) == "vol")
			cmd[1] = "dyn"			// compatibility
		set_pos(cmd[1], cmd[2])
		break
	case "sounding-score":
		cfmt.sound = "sounding"
		break
	case "staffwidth":
		v = get_unit(parm)
		if (isNaN(v)) {
			syntax(1, errs.bad_val, '%%' + cmd)
			break
		}
		if (v < 100) {
			syntax(1, "%%staffwidth too small")
			break
		}
		v = cfmt.pagewidth - v - cfmt.leftmargin
		if (v < 2) {
			syntax(1, "%%staffwidth too big")
			break
		}
		cfmt.rightmargin = v
		img.chg = 1 //true
		break
	case "textoption":
		cfmt[cmd] = get_textopt(parm)
		break
	case "dynalign":
	case "singleline":
	case "stretchlast":
	case "titletrim":
		v = +parm
		if (isNaN(v))
			v = get_bool(parm) ? 0 : 1
		if (cmd[1] == 't') {		// stretchlast
			if (v < 0 || v > 1) {
				syntax(1, errs.bad_val, '%%' + cmd)
				break
			}
		}
		cfmt[cmd] = v
		break
	case "combinevoices":
		syntax(1, "%%combinevoices is deprecated - use %%voicecombine instead")
		break
	case "voicemap":
		set_v_param("map", parm)
		break
	case "voicescale":
		set_v_param("scale", parm)
		break
	case "unsizedsvg":
		if (get_bool(parm))
			user.imagesize = ""
		else
			delete user.imagesize
		break
	default:		// memorize all global commands
//		if (typeof tm[cmd] == "function") {	// if in a module
//			tm[cmd](self)			// switch
//			self.set_format(cmd, parm)
//			break
//		}
		if (!parse.state)		// (needed for modules)
			cfmt[cmd] = parm
		break
	}

	// check if already a same format
	if (sfmt.indexOf(cmd) >= 0 && parse.ufmt) {
		// to do...
		parse.ufmt = 0 //false
	}
} // set_format()

// font stuff

// build a font style
function st_font(font) {
    var	n = font.name,
	r = ""

	if (font.weight)
		r += font.weight + " "
	if (font.style)
		r += font.style + " "
	if (n.indexOf('"') < 0 && n.indexOf(' ') > 0)
		n = '"' + n + '"'
	return r + font.size.toFixed(1) + 'px ' + n
} // st_font()
function style_font(font) {
	return 'font:' + st_font(font)
}
Mus.prototype.style_font = style_font

// build a font class
function font_class(font) {
//	if (font.cl)
//		return font.cl
    var	cl = 'f' + font.fx + cfmt.fullsvg

	if (font.class)
		cl += ' ' + font.class
	if (font.box)
		cl += ' box'
//	font.cl = cl
	return cl
} // font_class()

// use the font
function use_font(font) {
	if (font.fx == undefined) {	// if default font
		font.fx = get_fx(st_font(font))
		if (!font.swfac)
			set_font_fac(font)
		if (!font.pad)
			font.pad = 0

		// set the pointer to the width of the characters
		font.cw_tb = !font.name ? ssw_tb
			: font.name.indexOf("ans") > 0
				? ssw_tb		// sans-serif
				: font.name.indexOf("ono") > 0
					? mw_tb		// monospace
					: sw_tb		// serif
	}
	if (fxus[font.fx])
		return

    var	sel = ".f" + font.fx + cfmt.fullsvg
	fxus[font.fx] = 1 //true
	add_fstyle(sel + "{font:" + st_font(font) + "}")
	if (font.src)
		add_fstyle("@font-face{\n\
 font-family:" + font.name + ";\n\
 src:" + font.src + "}")
	if (font == cfmt.musicfont)	// add more music font style
		add_fstyle(sel + ' text,tspan{white-space:pre}')
} // use_font()

// get the font of the 'xxxfont' parameter
function get_font(fn) {
	fn += "font"
    var	font2, fx,
	font = cfmt[fn]

	if (!font) {
		syntax(1, "Unknown font $1", fn)
		return gene.curfont
	}

	if (!font.name || !font.size) {		// if incomplete user font
		font2 = clone(gene.deffont)
		if (font.name)
			font2.name = font.name
		if (font.normal) {
			if (font2.weight)	// !! don't use delete !!
				font2.weight = null
			if (font2.style)
				font2.style = null
		} else {
			if (font.weight)
				font2.weight = font.weight
			if (font.style)
				font2.style = font.style
		}
		if (font.src)
			font2.src = font.src
		if (font.size)
			font2.size = font.size
//		font2.fx = tm.styfx[st_font(font2)]
		font2.fx = undefined
		if (font.class)
			font2.class = font.class
		font = font2
	}
	use_font(font)
	return font
} // get_font()
