// txtmus - modules.js - module handling
//
// Copyright (C) 2018-2024 Jean-Francois Moine
//
// This file is part of txtmus-core.
//
// txtmus-core is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus-core is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus-core.  If not, see <http://www.gnu.org/licenses/>.

// empty function
if (!tm.loadjs) {
    tm.loadjs = function(fn, onsuccess, onerror) {
	if (onerror)
		onerror(fn)
	else
		onsuccess(fn)
    }
}

tm.modules = {
	alda: {},
	ambitus: {},
	autobar: {},
	begingrid: { fn: 'grid3' },
	beginps: { fn: 'psvg' },
	break: {},
	bww: {},
	capo: {},
	chordnames: {},
	clip: {},
	clairnote: { fn: 'clair' },
	cmni: {},
	cmno: {},
	voicecombine: { fn: 'combine' },
	diagram: { fn: 'diag' },
	equalbars: {},
	gamelan: {},
	grid: {},
	grid2: {},
	jazzchord: {},
	jianpu: {},
	mdnn: {},
	mei: {},
	MIDI: {},
	mnx: {},
	nns: {},
	pae: {},
	pageheight: { fn: 'page' },
	pedline: {},
	percmap: { fn: 'perc' },
	roman: {},
	soloffs: {},
	sth: {},
	strtab: {},
	temperament: { fn: 'temper' },
	tropt: {},

	nreq: 0,

	// scan the file and find the required modules
	// @file: text file
	// @relay: (optional) callback function for continuing the treatment
	// @errmsg: (optional) function to display an error message if any
	//	This function gets one argument: the message
	// return true when all modules are loaded
	load: function(file, relay, errmsg) {

		function get_errmsg() {
			if (typeof user == 'object' && user.errmsg)
				return user.errmsg
			if (typeof tm.printErr == 'function')
				return tm.printErr
			if (typeof alert == 'function')
				return function(m) { alert(m) }
			if (typeof console == 'object')
				return console.log
			return function(){}
		} // get_errmsg()

		// call back functions for loadjs()
		function load_end() {
			if (--tm.modules.nreq == 0)
				tm.modules.cbf()
		}

		// test if some keyword in the file
	    var	m, i, fn,
		nreq_i = this.nreq,
		ls = file.match(/(^|\n)(%%|I:).+?\b/g)

		if (file.indexOf('<mei') >= 0) {
			if (ls)
				ls.push('mei')
			else
				ls = ['mei']
		}
		if (file.indexOf('<mnx') >= 0) {
			if (ls)
				ls.push('mnx')
			else
				ls = ['mnx']
		}
		if (!ls)
			return 1 //true

		this.cbf = relay ||		// (only one callback function)
			function(){}
		this.errmsg = errmsg || get_errmsg()

		for (i = 0; i < ls.length; i++) {
			fn = ls[i].replace(/\n?(%%|I:)/, '')
			m = tm.modules[fn]
			if (!m || m.loaded)
				continue

			m.loaded = 1 //true	// load only once

			// load the module
			if (m.fn)
				fn = m.fn
			this.nreq++
			tm.loadjs(fn + "-2.js",
					load_end,
					function () {
						tm.modules.errmsg(
							'Error loading the module ' + fn)
						load_end()
					})
		}
		return this.nreq == nreq_i
	}
} // modules
