// ambitus.js - module to insert an ambitus at start of a voice
//
// Copyright (C) 2018-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%ambitus" appears in a ABC source.
//
// Parameters
//	%%ambitus 1

if (typeof tm == "undefined")
	tm = {}

tm.ambitus = function(mus) {

    function do_ambitus() {
    var	C = tm.C,
	s, v, p_v, min, max,
	v_tb = mus.get_v_tb()

	for (v = 0; v < v_tb.length; v++) {
		p_v = v_tb[v]
		if (p_v.second)
			continue
		min = 100
		max = -100

		// search the top and bottom pitches
		for (s = p_v.sym; s; s = s.next) {
			if (s.typ != C.note)
				continue
			if (s.notes[s.nhd].pit > max)
				max = s.notes[s.nhd].pit
			if (s.notes[0].pit < min)
				min = s.notes[0].pit
		}
		if (min == 100)
			continue			// no note

		s = p_v.clef
		s.stem = 1
		s.head = C.FULL
		s.stemless = true
		s.nhd = 1
		s.notes = [{
				dur: C.BLEN / 4,
				pit: min,
				shhd: 0
			},{
				dur: C.BLEN / 4,
				pit: max,
				shhd: 0
			}]
	}
    } // do_ambitus()

    function draw_symbols(of, p_voice) {
    var	staff_tb = mus.get_staff_tb(),
	s = p_voice.sym

	if (s.clef_type != undefined && s.nhd > 0) {
		s.x -= 26
		mus.set_scale(s)
		mus.draw_note(s)
		if (s.notes[1].pit - s.notes[0].pit > 4) {
			mus.xypath(s.x, 3 * s.notes[1].pit + staff_tb[s.st].y)
			mus.out_svg('v' +
				((s.notes[1].pit - s.notes[0].pit) * 3).toFixed(1) +
				'" stroke-width=".6"/>\n')
		}
		s.x += 26
		s.nhd = 0
		p_voice.clef.nhd = 0
	}
	of(p_voice)
    } // draw_symbols()

    function set_pitch(of, last_s) {
	of(last_s)
	if (!last_s && mus.cfmt().ambitus)
		do_ambitus()
    } // set_pitch()

    function set_glue(of, w) {
    var	v, s,
	v_tb = mus.get_v_tb()

	of(w)
	if (mus.cfmt().ambitus) {
		for (v = 0; v < v_tb.length; v++) {
			s = v_tb[v].sym
			while (!s.time && !s.clef_type)
				s = s.next
			if (s && s.nhd) {	// draw the helper lines if any
				s.x -= 26
				mus.draw_hl(s)
				s.x += 26
				s.ymn = 3 * s.notes[0].pit - 2
				s.ymx = 3 * s.notes[1].pit + 2
			}
		}
	}
    } // set_glue()

    function set_fmt(of, cmd, parm) {
	if (cmd == "ambitus")
		mus.cfmt().ambitus = parm
	else
		of(cmd, param)
    } // set_fmt()

    function set_width(of, s) {
	if (s.clef_type != undefined && s.nhd > 0) {
		s.wl = 40
		s.wr = 12
	} else {
		of(s)
	}
    } // set_width()

	// set the hooks
	mus.draw_symbols = draw_symbols.bind(mus, mus.draw_symbols)
	mus.set_pitch = set_pitch.bind(mus, mus.set_pitch)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.set_sym_glue = set_glue.bind(mus, mus.set_sym_glue)
	mus.set_width = set_width.bind(mus, mus.set_width)
} // ambitus

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.ambitus = tm.ambitus
