// autobar.js - module to add measure bars
//
// Copyright (C) 2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%autobar" appears in a text music source.
//
// Parameters
//	(none)

if (typeof tm == "undefined")
	tm = {}

tm.autobar = function(mus) {
    var	C = tm.C

    // hook on note creation
    function new_note(of, gr, sls) {
    var s2,
	tim = 0,
	curv = mus.get_curvoice(),
	s = of(gr, sls)

	if (!s)				// if error
		return s

	if (mus.cfmt().autobar) {
		if (curv.time < curv.wmeasure)
			return s		// still in the first measure
		s2 = curv.last_bar || curv.sym
		while (s2 != s) {
			if (s2.bar_type)
				tim = s2.time
			s2 = s2.next
		}
		do {
			tim += curv.wmeasure	// time of the next bar
		} while (tim < curv.time)
		if (tim == curv.time) {		// insert a bar
			s2 = {
				typ: C.bar,
				bar_type: "|",
				dur: 0
			}
			mus.sym_link(s2)
		}
		if (s2 != s)
			curv.last_bar = s2
	}
	return s
    } // new_note()

     function set_fmt(of, cmd, parm) {
	if (cmd == "autobar")
		mus.cfmt().autobar = mus.get_bool(parm)
	else
		of(cmd, parm)
    } // set_fmt()

	// set the hooks
	mus.new_note = new_note.bind(mus, mus.new_note)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
} // autobar()

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.autobar = tm.autobar
