// break.js - module to handle the %%break command
//
// Copyright (C) 2018-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%break" appears in a ABC source.
//
// Parameters
//	%%break measure_nb [":" num "/" den] [" " measure ...]*

if (typeof tm == "undefined")
	tm = {}

tm.break = function(mus) {

	// get the %%break parameters
	function get_break(parm) {
	    var	C = tm.C,
		b, c, d, sq,
		a = parm.split(/[ ,]/),
		cfmt = mus.cfmt()

		if (!cfmt.break)
			cfmt.break = []
		for (n = 0; n < a.length; n++) {
			b = a[n]
			c = b.match(/(\d+)([a-z]?)(:\d+\/\d+)?/)
			if (!c) {
				mus.syntax(1, mus.errs.bad_val, "%%break")
				continue
			}
			if (c[2])
				sq = c[2].charCodeAt(0) - 0x61
			if (!c[3]) {
				cfmt.break.push({	// on measure bar
						m: c[1],
						t: 0,
						sq: sq})
				continue
			}
			d = c[3].match(/:(\d+)\/(\d+)/)
			if (!d || d[2] < 1) {
				mus.syntax(1, "Bad denominator in %%break")
				continue
			}
			cfmt.break.push({
					m: c[1],
					t: d[1] * C.BLEN / d[2],
					sq: sq})
		}
	} // get_break()

	// insert the EOLs of %%break
	function do_break() {
	    var	i, m, t, brk, seq,
		v_tb = mus.get_v_tb()
		v = mus.get_cur_sy().top_voice,
		s1 = v_tb[v].sym,
		cfmt = mus.cfmt()

		for (i = 0; i < cfmt.break.length; i++) {
			brk = cfmt.break[i]
			m = brk.m
			for (s = s1; s; s = s.next) {
				if (s.bar_n == m)
					break
			}
			if (!s)
				continue

			if (brk.sq) {
				seq = brk.sq
				for (s = s.ts_next; s; s = s.ts_next) {
					if (s.bar_n == m) {
						if (--seq == 0)
							break
					}
				}
				if (!s)
					continue
			}

			t = brk.t
			if (t) {
				t = s.time + t
				for ( ; s; s = s.next) {
					if (s.time >= t)
						break
				}
			} else {
				s = s.next
			}
			if (s)
				s.soln = true
		}
	} // do_break()

	function set_fmt(of, cmd, parm) {
		if (cmd == "break")
			get_break(parm)
		else
			of(cmd, parm)
	} // set_fmt()

	function set_bar_num(of) {
		of()
		if (mus.cfmt().break)
			do_break()
	} // set_bar_num()

	// set the hooks
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.set_bar_num = set_bar_num.bind(mus, mus.set_bar_num)
} // break

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.break = tm.break
