// capo.js - module to add a capo chord line
//
// Copyright (C) 2018-2022 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%capo" appears in a ABC source.
//
// Parameters
//	%%capo n	'n' is the capo fret number

if (typeof tm == "undefined")
	tm = {}

tm.capo = function(mus) {

	// b40 intervals of capo
    var	icb40 = [0, 5, 6,11,16,17,22,23,28,33,34,39]
//		 e  f ^f  g _a  a _b  b  c _d  d _e

    function gch_build(of, s) {
    var	t, i, gch, gch2, i2,
	p_v = mus.get_curvoice(),
	a_gch = s.a_gch

	if (p_v.capo && a_gch) {
		t = p_v.capo
		i = 0

		while (1) {
			gch = a_gch[i++]
			if (!gch) {
				of(s)
				return
			}
			if (gch.type == 'g')
				break
		}
//		gch2 = Object.create(gch)
		gch2 = {
			type: 'g',
			istart: gch.istart,
			iend: gch.iend,
			pos: gch.pos
		}
		gch2.capo = 0 //false	// (would be erased when setting gch)
		gch2.t = mus.gch_tr1(gch.t, -icb40[t % 12])
		if (!p_v.capo_first) {		// if new voice
			p_v.capo_first = 1 //true
			gch2.t += "  (capo: " + t.toString() + ")"
		}

		gch2.f = mus.get_font(mus.cfmt().capofont ?
						"capo" : "annotation")
		a_gch.splice(i, 0, gch2)

		// set a mark in the first chord symbol for %%diagram
		gch.capo = 1 //true
	}
	of(s)
    }

    function set_fmt(of, cmd, parm) {
	if (cmd == "capo")
		mus.set_v_param("capo_", parm)
	else
		of(cmd, parm)
    } // set_fmt()

    // get the parameters of the current voice
    function set_vp(of, a) {
    var	v,
	p_v = mus.get_curvoice(),
	i = 0

	while (1) {
		if (!a[i])
			break
		if (a[i++] == "capo_"
		 && a[i] == '=') {
			v = +a[++i]
			if (isNaN(v) || v <= 0)
				mus.syntax(1, "Bad fret number in %%capo")
			else
				p_v.capo = v
			break
		}
	}
	of(a)
    } // set_vp()

	// set the hooks
	mus.gch_build = gch_build.bind(mus, mus.gch_build)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.set_vp = set_vp.bind(mus, mus.set_vp)
} // capo

// the module is loaded
if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.capo = tm.capo
