// txtmus - chordnames.js - change the names of the chord symbols
//
// Copyright (C) 2020-2023 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded by %%chordnames.
//
// Parameters
// 1st syntax:
//		%%chordnames <comma separated list of chord names>
//	Each name replace one chord. The order is:
//		CDEFGAB<N.C.>
// 2nd syntax:
//		%%chordnames <comma separated list of key ':' value>
//	The key may be a chord letter ('A') with/or an accidental
//	Example:
//		%%chordnames Bb:B,B:H,b:s	% German chords

if (typeof tm == "undefined")
	tm = {}

tm.chordnames = function(mus) {

    function gch_build(of, s) {
    var	gch, ix, t,
	cfmt = mus.cfmt()

	if (s.a_gch && cfmt.chordnames) {
		for (ix = 0; ix < s.a_gch.length; ix++) {
			gch = s.a_gch[ix]
			t = gch.t
			if (gch.type != 'g' || !t)
				continue
			if (t[0] == 'n' || t[0] == 'N')
				t = 'N'
			gch.t = t.replace(cfmt.chordnames.re,
				function(c){return cfmt.chordnames.o[c]})
		}
	}
	of(s)
    } // gch_build()

    var gimpl = 'CDEFGABN'

    function set_fmt(of, cmd, parm) {
    var	i, v,
	re = [],
	o = {},
	cfmt = mus.cfmt()

	if (cmd == "chordnames") {
		v = parm.split(/[\s=:,]+/)
		if (parm.indexOf(':') > 0
		 || parm.indexOf('=') > 0) {
			for (i = 0; i < v.length; i += 2) {
				o[v[i]] = v[i + 1]
				re.push(v[i])
			}
		} else {			// implicit
			for (i = 0; i < v.length; i++) {
				o[gimpl[i]] = v[i]
				re.push(gimpl[i])
			}
		}
		cfmt.chordnames = {re: new RegExp(re.join('|'), 'g'), o: o}
		return
	}
	of(cmd, parm)
    } // set_fmt()

	// set the hooks
	mus.gch_build = gch_build.bind(mus, mus.gch_build)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
} // chordnames

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.chordnames = tm.chordnames
