// cmno.js - module to output the Carnatic Music Notation
//
// Copyright (C) 2022-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%cmno" appears in a ABC source.
//
// Parameters
//	%%cmno [options]
//

if (typeof tm == "undefined")
	tm = {}

tm.cmno = function(mus) {

    var	cgd2cde = new Int8Array([0, -4, -1, -5, -2, -6, -3,
				 0, -4, -1, -5, -2, -6, -3, 0]),
	cc_tb = ["\ue264", "\ue260", , "\ue262", "\ue263", "\ue261"]

// change %%staves and %%score
    function do_pscom(of, p) {
	switch (p.match(/\w+/)[0]) {
	case 'staves':
	case 'score':
		p = p.replace(/\(|\)/g, '')
		break
	}
	of(p)
  } // do_pscom()

// adjust some symbols before the generation
  function output_music(of) {
    var	C = tm.C,
	s2, s3,
	cur_sy = mus.get_cur_sy(),
	v_tb = mus.get_v_tb()

	if (!mus.cfmt().cmno) {
		of()
		return
	}

	function set_sym(p_v) {
	    var s, s2, note, pit, nn, p, a, m, i,
		sf = p_v.key.k_sf,
		delta = cgd2cde[sf + 7] + 16 // = 2 + 14

		p_v.key.k_a_acc = []	// no accidental

		// no (visible) clef
		s = p_v.clef
		s.invis = 1 //true
		s.clef_type = 't'
		s.clef_line = 2

		// scan the voice
		for (s = p_v.sym; s; s = s.next) {
			s.st = p_v.st
			switch (s.typ) {
			case C.clef:
				s.invis = 1 //true
				s.clef_type = 't'
				s.clef_line = 2
//				continue
			default:
				continue
			case C.key:
				delta = cgd2cde[s.k_sf + 7] - 2
				s.invis = 1 //true
				continue
			case C.rest:
				note = s.notes[0]
				if (s.prev && s.prev.typ == C.note)
					note.jn = '-,'
				else
					note.jn = ','
				delete s.fmr
				s.typ = C.note
				s.stemless = 1 //true
				note.pit = 7		// "e"
				note.jo = 2		// octave number
				continue
			case C.note:			// change the notes
				break
			}

			s.stem = -1
			s.stemless = 1 //true

			if (s.sls) {
				for (i = 0; i < s.sls.length; i++)
					s.sls[i].ty = C.SL_ABOVE
			}

//			for (m = 0; m <= s.nhd; m++) {
//				note = s.notes[m]
			note = s.notes[0]

				// note head
				p = note.pit
				pit = p + delta
				note.jn = "srgmpdn"[(pit + 77) % 7]

				// set a fixed offset to the note for the decorations
				note.pit = 7			// "e"

				note.jo = (pit / 7) | 0	// octave number


				// set the slurs and ties up
				if (note.sls) {
					for (i = 0; i < note.sls.length; i++)
						note.sls[i].ty = C.SL_ABOVE
				}
				if (note.tie_ty)
					note.tie_ty = C.SL_ABOVE
//			}

//			// replace the staccato dot
//			if (s.a_dd) {
//				for (i = 0; i < s.a_dd.length; i++) {
//					if (s.a_dd[i].glyph == "stc") {
//						mus.deco_put("gstc", s)
//						s.a_dd[i] = s.a_dd.pop()
//					}
//				}
//			}
		}
	} // set_sym()

	// -- output_music --

	for (v = 0; v < v_tb.length; v++)
		set_sym(v_tb[v])

	of()
  } // output_music()

  function draw_symbols(of, p_voice) {
    var	i, m, nl, note, s, s2, x, y,
	C = tm.C,
	dot = "\ue1e7",
	staff_tb = mus.get_staff_tb(),
	out_svg = mus.out_svg,
	out_sxsy = mus.out_sxsy,
	xypath = mus.xypath

	if (!mus.cfmt().cmno) {
		of(p_voice)
		return
	}

	// draw the duration lines under the notes
	function draw_dur(s1, y, s2, n, nl) {
	    var s, s3

		xypath(s1.x - 3, y + 3)
		out_svg('h' + (s2.x - s1.x + 8).toFixed(1) + '"/>\n')	// "
		y -= 3
		while (++n <= nl) {
			s = s1
			while (1) {
				if (s.nflags && s.nflags >= n) {
					s3 = s
					while (s != s2) {
						if (s.next.bmbrk && n > s.next.bmbrk
						 || (s.next.nflags
						  && s.next.nflags < n))
							break
						s = s.next
					}
					draw_dur(s3, y, s, n, nl)
				}
				if (s == s2)
					break
				s = s.next
			}
		}
	} // draw_dur()

	function out_mus(x, y, p) {
		out_svg('<text x="')
		out_sxsy(x, '" y="', y)
		out_svg('">' + p + '</text>\n')
	} // out_mus()

	function out_txt(x, y, p) {
		out_svg('<text class="bj" x="')
		out_sxsy(x, '" y="', y)
		out_svg('">' + p + '</text>\n')
	} // out_txt()

	function draw_hd(s, x, y) {
	    var	m, note, ym, i, c

		note = s.notes[0]
		c = note.jn
			out_txt(x - 3.5, y + 8, c)
			if (note.jo > 2) {
				out_mus(x - 1, y + 22, dot)
				if (note.jo > 3) {
					y += 3
					out_mus(x - 1, y + 22, dot)
				}
			} else if (note.jo < 2) {
				ym = y + 4
				if (m == 0 && s.nflags > 0)
					ym -= 3 * s.nflags
				out_mus(x - 1, ym, dot)
				if (note.jo < 1) {
					ym -= 3
					out_mus(x - 1, ym, dot)
				}
			}
	} // draw_hd()

	// -- draw_symbols --

	for (s = p_voice.sym; s; s = s.next) {
		if (s.invis)
			continue
		switch (s.typ) {
//		case C.meter:
//			mus.draw_meter(s)
//			break
		case C.note:
		case C.rest:
			x = s.x
			y = staff_tb[s.st].y
			draw_hd(s, x, y)

			if (s.nflags >= 0 && s.dots)
				out_mus(x + 8, y + 13, dot)
			if (s.nflags > 0) {
//fixme: ko with rests because no beam_st /_end
				if (s.beam_st || s.typ == C.rest) {
					nl = s.nflags
					s2 = s
					while (1) {
						if (s2.nflags && s2.nflags > nl)
							nl = s2.nflags
						if (s2.beam_end)
							break
						if (!s2.next
						 || !s2.next.nflags
						 || s2.next.nflags <= 0)
							break
						s2 = s2.next
					}
					draw_dur(s, y, s2, 1, nl)
				}
			}
			break
		}
	}
    } // draw_symbols()

// set some parameters
    function set_fmt(of, cmd, parm) {
	if (cmd == "cmno") {
	    var	cfmt = mus.cfmt()

		cfmt.cmno = mus.get_bool(parm)
		if (!cfmt.cmno)
			return
		cfmt.staffsep = 20
		cfmt.sysstaffsep = 14
		mus.param_set_font("vocalfont", "monospace 15")
		mus.set_v_param("stafflines", "...")
		cfmt.tuplets = [0, 1, 0, 1]	// [auto, slur, number, above]
		cfmt.stretchstaff = 0
		cfmt.stretchlast = 0
		cfmt.spc_tb = new Float32Array([
			20,20,20,20,20,
			20,
			20,20,20,20])
		return
	}
	of(cmd, parm)
    } // set_fmt()

// adjust some values
    function set_pitch(of, last_s) {
	of(last_s)
	if (!last_s
	 || !mus.cfmt().cmno)
		return			// first time
    var	C = tm.C
	
	for (var s = mus.get_tsfirst(); s; s = s.ts_next) {
		switch (s.typ) {

//		// draw the key signature only in the first voice
//		// and not at start of the staff
//		case C.key:
//			if (s.prev.typ == C.clef
//			 || s.v != 0)
//				s.a_gch = null
//			break

		// adjust the vertical spacing above the note heads
		case C.note:
			s.ymx = 20 * s.nhd + 22
			if (s.notes[s.nhd].jo > 2) {
				s.ymx += 3
				if (s.notes[s.nhd].jo > 3)
					s.ymx += 2
			}
			break
		}
	}
    } // set_pitch()

// set the width of some symbols
    function set_width(of, s) {
	if (s.wl != undefined)
		return			// already computed
	of(s)
	if (!mus.cfmt().cmno)
		return

    var	w, i, c, //m, note,
	C = tm.C

	switch (s.typ) {
	case C.clef:
		s.wl = s.wr = .1	// !! not 0 !!
		break
	case C.key:
	case C.meter:
		s.wl = s.wr = 0
		break
	case C.note:
	case C.rest:
//		c = s.notes[0].jn
		i = 0
		if (s.dur > C.BLEN)
			i++
		if (s.dur > C.BLEN / 4 * 3)
			i++
		if (s.dur > C.BLEN / 2)
			i++
		if (s.dur > C.BLEN / 4)
			i++
//console.log(s.time+' jn: '+s.notes[0].jn+' '+",,,,".slice(0, i)+' i: '+i)
		if (i)
			s.notes[0].jn += ",,,,".slice(0, i)

		w = 6 + 12 * i
		s.wl = 0
		s.wr = w

		if (s.notes[0].jn[0] == '-')	// start of rest
			s.wr += 12

//		s.nflags = 0
		s.dots = 0
		break
	}
    } // set_width()

	// set the hooks
	mus.do_pscom = do_pscom.bind(mus, mus.do_pscom)
	mus.draw_symbols = draw_symbols.bind(mus, mus.draw_symbols)
	mus.output_music = output_music.bind(mus, mus.output_music)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.set_pitch = set_pitch.bind(mus, mus.set_pitch)
	mus.set_width = set_width.bind(mus, mus.set_width)

	// big staccato dot
	mus.glyphs.gstc = '<circle id="gstc" cx="0" cy="-3" r="2"/>'
	mus.decos.gstc = "0 gstc 5 1 1"

	mus.add_style("\n.bj {font-family:monospace; font-size:15px}")
} // cmno

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.cmno = tm.cmno
