// grid2.js - module to replace a voice in the music by a chord grid
//
// Copyright (C) 2018-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%grid2" appears in a ABC source.
//
// Parameters
//	%%grid2 y
// This command must appear globally or in a voice.

if (typeof tm == "undefined")
	tm = {}

tm.grid2 = function(mus) {

// function called before tune generation
    function do_grid() {
    var s, v, p_v, ix, cs, c_a_cs, bt, gch,
	v_tb = mus.get_v_tb()

	if (mus.cfmt().grid2)
		for (v = 0; v < v_tb.length; v++)
			if (v_tb[v].grid2 == undefined)
				v_tb[v].grid2 = 1

	for (v = 0; v < v_tb.length; v++) {
		p_v = v_tb[v]
		if (!p_v.grid2)
			continue
		curvoice = p_v
		mus.set_v_param("stafflines", "...")	// no staff
		p_v.clef.invis = 1 //true		// no clef
		p_v.key.k_sf = 0			// no key signature
		delete p_v.key.k_a_acc
		p_v.staffnonote = 2			// draw the line
		for (s = p_v.sym; s; s = s.next) {
			delete s.a_dd			// no decoration
			if (!s.dur) {
				if (s.bar_type)
					bt = s.time
				continue
			}

			// set all notes
			s.invis = true	//  as invisible
			delete s.sl1	//  with no slur
			delete s.sls
			delete s.tie_s	//  and no tie
			delete s.ti2
			for (ix = 0; ix <= s.nhd; ix++) {
				delete s.notes[ix].tie_e
				delete s.notes[ix].tie_ty
			}
			if (s.tf)	// don't show the tuplets
				s.tf[0] = 1
			if (!s.a_gch) {
				if (s.time == bt)
					s.a_gch = [ mus.clone(c_a_cs) ]
				continue
			}
			for (ix = 0; ix < s.a_gch.length; ix++) {
				gch = s.a_gch[ix]
				if (gch.type == 'g') {
					c_a_cs = gch
					break
				}
			}
		}
	}
    } // do_grid()

    // draw the chord symbol in the middle of the staff
    function draw_gchord(of, i, s, x, y) {
    var	csan

	if (s.p_v.grid2) {
		mus.set_dscale(s.st)
		csan = s.a_gch[i]
		if (csan.type == 'g')
			mus.xy_str(s.x + csan.x, 12 - csan.f.size * .5,
					csan)
	} else {
		of(i, s, x, y)
	}
    } // draw_gchord()

    function output_music(of) {
	do_grid()
	of()
    } // output_music()

    function set_fmt(of, cmd, param) {
	if (cmd == "grid2") {
	    var	curvoice = mus.get_curvoice(),
		v = mus.get_bool(param)

		if (curvoice)
			curvoice.grid2 = v
		else
			mus.cfmt().grid2 = v
		return
	}
	of(cmd, param)
    } // set_fmt()

	// set the hooks
	mus.draw_gchord = draw_gchord.bind(mus, mus.draw_gchord)
	mus.output_music = output_music.bind(mus, mus.output_music)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
} // grid2

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.grid2 = tm.grid2
