// txtmus - mnx.js - MNX front-end
//
// Copyright (C) 2022-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.

if (typeof tm == "undefined")
	tm = {}

tm.mnx = function(mus) {
    // convert MNX to music suitable to the txtmus generator
    // This function is called from the ABC frontend.
    tm.mnx.mnx2mus = function (	fname,		// file name
				mnx,		// file content
				bof,
				eof) {
    var	errs = {
		bad_pit: "Bad note pitch"
	},
	k_bar = {			// bars
		dashed: ".",
		dotted: ".",
		heavy: "]",
		"heavy-heavy": "]]",
		"heavy-light": "]|",
		"light-heavy": "|]",
		"light-light": "||",
		none: "[]",
		regular: "|",
		short: "|",
		tick: "|"
	}, // k_bar
	syfl = {			// system flags
		o1_brace: tm.C.O_BRACE,
		o1_bracket: tm.C.O_BRACKET,
		o2_brace: tm.C.O2_BRACE,
		o2_bracket: tm.C.O2_BRACKET,
		c1_brace: tm.C.C_BRACE,
		c1_bracket: tm.C.C_BRACKET,
		c2_brace: tm.C.C2_BRACE,
		c2_bracket: tm.C.C2_BRACKET
	}, // syfl

	k_tb = [			// key names
		"Cb", "Gb", "Db", "Ab", "Eb", "Bb", "F",
		"C",
		"G", "D", "A", "E", "B", "F#", "C#", "G#", "D#", "A#"
	] // k_tb

	if (bof == undefined)
		bof = 0
	if (!eof)
		eof = mnx.length

//    var	i = mnx.indexOf('<mnx', bof)
//	if (i < 0 || i > eof) {
//		tm.mnx.tosvg.call(mus, fname, mnx, bof, eof)
//		return
//	}
//	if (i > bof) {
//		tm.mnx.tosvg.call(mus, fname, mnx, bof, i)
//		bof = i
//	}

    var	cfmt = mus.cfmt(),		// parameters
	parse = mus.parse,		// parse variables
	C = tm.C,			// constants
	g = {				// global values
//		bh: {},		// <beam-hooks>
		bm: {},		// beam start level/stop - index = event ID
//		bml: 0,		// beam level
		d: [],		// <directions>
//		dee: [],	// decoration end as event ID
//		p_v: {},	// pointer to the first voice of the part
//		grace: 0,	// inside a <grace> note sequence
		grll: 0,	// <group-layout> level
		meas: [],	// <global> definition of the measures
//		mn: -1,		// current measure number - reset on <part> start
//		nid: {},	// note IDs
//		p: {},		// <directions-part>
//		part:		// tag of the current <part>
//				//	plus vo[] = array of voices
//		score: '',	// selected  <score>
//		sl: {},		// <slur> ending on an event
//		sln: [],	// <slur> ending on a note
		sysl: {},	// system-layout.id -> staff system
				//	staves: flags
				//	voices:
				//		st: staff number (0..n-1)
				//		part: part ID
				//		pst: staff string in part
				//		pvo: voice string in part
//		syst: [],	// chosen staff system = array of staff systems
//				//	staves: flags
//				//	voices: (id sysl)
//				//	loc: start measure/location
//		systx:		// staff system index in syst[]
//		ti: null,	// <tied>
//		tim: 0,		// start time of the current measure
//		v: 0,		// current voice number in the part
		vo: 1		// voice ID when no part/voice ID
	},

	// --- functions associated to the start of MNX tags
    fns = {

	// inside beams
	beam: function(tag) {
	    var	a = tag.events.split(' '),	// get the first and last events
		e = a[0]			// first event

		if (!g.bml)
			g.bm[e] = 0		// stop the previous beam if any

		e = a[a.length - 1]		// last event
		g.bml++				// beam level
		if (!g.bm[e])			// keep the lowest level
			g.bm[e] = g.bml		// stop current beam
	}, //beam

	// inside <beam>
	"beam-hook": function(tag) {
		if (tag.direction[0] == 'l') {	// ignore the "right"
			if (!g.bh)
				g.bh = {}
			g.bh[tag.event] = 1
		}
	}, // beam-hook

	// inside <sequence>
	beams: function(tag) {
		g.bml = 0	// beam level
	}, // beams

	// inside <directions-part> or <directions>
	clef: function(tag) {
		if (!g.p) {			// in directions
			mus.set_vp([ "clef", '=',
				(tag.sign + tag.line
					+ (tag.octave
						? (tag.octave * 7 + 1).toString()
						: '')) ])
		} else {			// in directions-part
			g.p.clef = tag
		}
	}, // clef

	// inside <directions>
	cresc: function(tag) {
		g.d.push(tag)
	}, // cresc

//	// inside <sequence>
//	directions: function(tag) {
//	}, // directions

//	// inside <measure-global>
//	"directions-global": function(tag) {
//	}, // directions-global

	// inside <measure>
	"directions-part": function(tag) {
		g.p = {}
	}, // directions-part

	// inside <directions-global>
	ending: function(tag) {
	    var	m = g.meas[g.meas.length - 1]

		if (tag.type == "start")
			m.end_st = tag
		else
			m.end_end = tag
	}, // ending

	// inside <sequence>
	// .color
	// .duration (as num/den'd' part of whole note) real duration
	// .id
	// .measure	yes/no (full measure rest)
	// .orient	up/down (of what?)
	// .smufl-font font name
	// .staff	(for cross staff voices)
	// .stem-direction up/down
	// .value	duration (not needed if tag.measure) = display value
	event: function(tag) {
	    var	d, ev, ev1, i,
		p_v = mus.get_curvoice(),
		s = {
			fname: parse.fname,
			istart: tag.ix,
			stem: 0,
			multi: 0,
			nhd: 0,
			xmx: 0
		}

		if (g.grace)
			s.grace = 1

		d = durcnv(tag.value)
		if (!d) {
			mus.error(1, s, 'Bad note/rest value ' + tag.value)
			return
		}
		s.dur = s.dur_orig = d

//--fixme: gerer aussi ici un possible new staff system
		if (p_v.de && p_v.de.length) {	// if some end of decoration
			for (i = 0; i < p_v.de.length; i++) {
				ev = p_v.de[i]	// [ #measure, time, deco ]
				if (ev[0] != g.mn + 1)
					continue
				if (ev[1] + g.tim < p_v.time) {
					if (typeof ev[2] == "number")
						s.ottava = [ev[2]] // octave-shift
					else
						mus.dcn_add(ev[1])
					p_v.de.splice(i, 1)
					i--
				}
			}
		}

		if (tag.id) {
			ev = g.bh && g.bh[tag.id]	// beam hooks
			if (ev) {
				switch (d) {
				case 96:		// 1/16th note
					mus.dcn_add('beambr1')
					break
				case 48:		// 1/32th note
					mus.dcn_add('beambr2')
					break
				}
				delete g.bh[tag.id]
			}
			ev = g.dee && g.dee[tag.id]	// delayed decorations
			if (ev) {
				while (1) {
					ev1 = ev.shift()
					if (!ev1)
						break
					mus.dcn_add(ev1)
				}
				delete g.dee[tag.id]
			}
		}

		if (g.d.length) {	// handle the <directions>
			while (1) {
				ev = g.d.shift()
				if (!ev)
					break
				switch (ev._nm) {
//				case "clef":	// directly treated
//				case "cresc":
//fixme
//					break
//				case "dim":
//fixme
//					break
				case "dynamics":
//fixme: staff
					switch (tag.orient) {
					case "up":
						s.pos.vol = C.SL_ABOVE
						break
					case "down":
						s.pos.vol = C.SL_BELOW
						break
					}
					mus.dcn_add(ev.type)
					break
//				case "expressions":
//fixme
//					break
//				case "instructions":
//fixme
//					break
				case "octave-shift":
//fixme: orient, staff
					i = 0
					switch (ev.type) {
					case "-15": i = 2; break
					case "-8": i = 1; break
					case "8": i = -1; break
					case "15": i = -2; break
					}
					if (!i)
						break
//fixme many octave-shift's?
					s.ottava = [i]		// delta octave
					de_add(ev.end, 0)	// ottava end
					mus.glovar().ottava = 1 // generate the deco
								// and shift
					break
				case "wedge":
//fixme: staff
					switch (tag.orient) {
					case "up":
						s.pos.dyn = C.SL_ABOVE
						break
					case "down":
						s.pos.dyn = C.SL_BELOW
						break
					}
					mus.dcn_add(ev.type + "(")
					de_add(ev.end, ev.type + ")")
					break
				}
			}
		}

		mus.sym_link(s)			// link inside voice

		if (tag.color)
			s.color = tag.color

		p_v.last_note = s
		if (g.tpl) {
			g.tpl.r++		// number of notes in the tuplet
//fixme: do the nested tuplets
			if (!g.tpl.s) {
				s.tp = [ g.tpl ]
				g.tpl.s = s	// first note of the tuplet
			}
		}
	}, //event

	// inside <directions-global>
	fine: function(tag) {
		g.meas[g.meas.length - 1].fine = tag
	}, // fine

	// inside <sequence> or <tuplet>
	forward: function(tag) {
	    var	s = {
			typ: C.rest,
			fname: parse.fname,
			istart: tag.ix,
			stem: 0,
			multi: 0,
			nhd: 0,
			xmx: 0,
			invis: 1
		},
		d = durcnv(tag.duration)

		if (!d) {
			mus.error(1, s, 'Bad forward duration ' + tag.duration)
			return
		}
		mus.sym_link(s)
	}, // forward

	// inside <mnx>
//	global: function(tag) {
//	}, // global

	// inside <sequence>
	grace: function(tag) {
	    var	s = {
			typ: C.grace,
			fname: parse.fname,
			istart: tag.ix,
			dur: 0,
			multi: 0
		}

		mus.sym_link(s)
		g.grace = s
		if (tag.slash != "no")
			s.sappo = 1	// acciaccatura
//fixme: "type" is used for playback
	}, // grace

	// inside <system-layout> or <group-layout>
	// O tag.label	name of the group on the left of the staff
	// O tag.symbol	(brace bracket line none square)
	"group-layout": function(tag) {
		g.grll++		// level

	    var	fl = syfl['o' + g.grll + '_' + tag.symbol],
		st = g.sy.nstaff

//console.log('group '+st+' '+tag.symbol+' '+fl)
		if (fl)
			g.sy.staves[st] |= fl
	}, // group-layout

	// inside <directions-global>
	jump: function(tag) {
		g.meas[g.meas.length - 1].jump = tag
	}, // jump

	// inside <directions-global> or <directions-part>
	key: function(tag) {
		if (!g.p)			// in directions-global
			g.meas[g.meas.length - 1].key = tag
		else				// in directions-part
			g.p.key = tag
	}, // key

//	// inside <mnx>
//	layouts: function(tag) {
//	}, // layouts

	// inside <part>
	measure: function(tag) {
	    var	p_v = mus.get_curvoice()

		g.v = 0			// voice restart
		g.mn++
		g.tim = p_v.time
		g.p = null		// no direction yet
	}, // measure

	// inside <global>
	// .barline O
	// .index O		// measure index
	// .number O		// visual measure label
	"measure-global": function(tag) {
		g.meas.push(tag)	// keep the definition of the measures
	}, // measure-global

	// root
	mnx: function(tag) {
		mus.info.X = "1"
		parse.fname = fname,
		parse.state = 1		// tune header
	}, // mnx

	// inside <event>
	note: function(tag) {
	    var	r, m, not, pit, a, i, not1, ev, sl,
		af = 0,
		p_v = mus.get_curvoice(),
		s = p_v.last_sym	// event

		s.typ = C.note

		r = tag.pitch.match(/([A-G])([#b]*)(\d)(.*)/)
		if (!r) {
			mus.error(1, s, errs.bad_pit)
			return
		}
		pit = "CDEFGAB".indexOf(r[1]) + (+r[3] * 7) - 28
		switch (r[2]) {
		case 'bb': a = -2; break
		case 'b': a = -1; break
		case '#': a = 1; break
		case '##': a = 2; break
		default: a = 0; break
		}
		if (r[4]) {
			if (r[4][0] == '+') {
				af = 1
			} else if (r[4][0] == '-') {
				af = -1
			} else {
				mus.error(1, s, errs.bad_pit)
				return
			}
			af *= Number(r[4])
			switch (r[4].slice(-1)) {
			case 'o':
				af *= 12
				break
			case 'w':
				af *= 2
				break
			default:
				i = r[4].indexOf('/')
				if (i > 0)
					af /= r[4].slice(i)
				break
			}
		}
//		mid = mus.pit2mid(pit + 19, a + af)

		if (!s.notes) {
			s.notes = []
			m = 0
		} else {
			m = s.notes.length
		}
		s.nhd = m
		not = s.notes[m] = {
			pit: pit,
			acc: af,
//			midi: mid,
			shhd: 0,
			shac: 0,
			dur: s.dur
		}

		switch (tag.accidental) {
		case "dbleflat": not.acc -= 2; break
		case "flat": not.acc--; break
		case "natural": not.acc = 3; break
		case "sharp": not.acc++; break
		case "dblesharp": not.acc += 2; break
		}
		mus.set_midi(not)

		if (tag.visibility == 'hidden')
			s.invis = 1
		if (tag.color)
			not.color = tag.color

		mus.deco_cnv(s)

		if (tag.id) {
			not1 = g.ti && g.ti[tag.id]
			if (not1) {			// tie
				not1.tie_ty = C.SL_AUTO
				not1.tie_e = not
				not1.s.ti1 = 1
				not.tie_s = not1
				not.s = s
				s.ti2 = 1
			}

			ev = g.sln && g.sln[tag.id]
			if (ev) {			// slur
				while (1) {
					sl = ev.shift()	// ({ty, ss})
					if (!sl)
						break
					if (sl.ss == s)
						sl.nts = not	// start note
					else
						sl.nte = not	// end note
				}
				delete g.sln[tag.id]
			}

			// memorize the ID in case a slur starts on the note
			if (!g.nid)
				g.nid = {}
			g.nid[tag.id] = not
		}
	}, // note

//	// inside <score>
//	page: function(tag) {
//// tag
////	layout	system-layout for this page
//	}, // page

	// inside <mnx>
	// tag.id:
	// tag.staves:
	part: function(tag) {
	    var	i, gsy, csy

		if (parse.state < 2) {		// if in the tune header
			mus.get_key('C')	// start the tune body
			if (g.syst) {		// if score
				g.systx = 0	// first staff system
				gsy = g.syst[0]	// update the first system
				csy = mus.get_cur_sy()
				csy.nstaff = gsy.nstaff
			}
		}
		g.part = tag		// keep the part data
		g.part.vo = []		// voices of the part
		g.part.nst = tag.staves || 1 // number of staves
		g.mn = -1		// current measure number
	}, // part

	// inside <staff-layout> or <voice-layout>
	// M tag.part
	// O tag.part-staff
	// O tag.part-voice
	// O tag.staff (up/down)
	"part-layout": function(tag) {
	    var	vo = {
			st: g.sy.nstaff,
			part: tag.part
		}

		if (tag["part-staff"])
			vo.pst = tag["part-staff"]
		if (tag["part-voice"])
			vo.pvo = tag["part-voice"]
//console.log('part-layout '+tag.part+' '+g.sy.voices.length
//	+(vo.pst?(' st '+vo.pst):'')
//	+(vo.pvo?(' vo '+vo.pvo):''))

		g.sy.voices.push(vo)
	}, // part-layout

	// inside <directions-global>
	repeat: function(tag) {

		// there may be two repeats in a measure |: .. :|
		if (tag.type == "start")
			g.meas[g.meas.length - 1].repl = tag.type
		else
			g.meas[g.meas.length - 1].repr = tag.type
	}, // repeat

	rest: function(tag) {
	    var	p_v = mus.get_curvoice(),
		s = p_v.last_sym		// event

		s.typ = C.rest

		if (s.dur == s.p_v.wmeasure)
			s.fmr = 1		// full measure rest
		if (tag.visibility == 'hidden')
			s.invis = 1
		if (tag.color)
			s.color = tag.color

		mus.deco_cnv(s)
	}, // rest

	// inside <mnx>
	score: function(tag) {
		if (g.score
		 || (cfmt.render
		  && cfmt.render != tag.name))
			return 1		// skip

		// only the selected or first score is rendered
		g.score = cfmt.render || tag.name
// tag.layout	ID of the system-layout
	}, // score

	// inside <directions-global>
	segno: function(tag) {
		g.meas[g.meas.length - 1].segno = tag
	}, // segno

	// inside <measure>
	// tag.staff: staff number ("1" .. "n")
	// tag.voice: voice ID
	sequence: function(tag) {
	    var	s, t, p_v, v, i, id, gsy, csy, p_gv,
		v_tb = mus.get_v_tb(),
		val = '',
		m = g.meas[g.mn]		// measure global definition

		if (g.syst) {			// if score
			gsy = g.syst[g.systx]

			// search the part and (staff or voice)
			// if the voice is not defined yet
			if (g.part.vo[g.v] == undefined) {
				id = g.part.id
//console.log('part '+id+' v:'+g.v
// +(tag.staff?(' st: '+tag.staff):'')
// +(tag.voice?(' vo: '+tag.voice):''))
				for (i = 0; i < gsy.voices.length; i++) {
					p_gv = gsy.voices[i]
					if (p_gv.part == id) {
//console.log(' '+i+' 1st '+p_gv.pst+' vo '+p_gv.pvo)
						if (p_gv.pst && tag.staff != p_gv.pst)
							continue
						if (p_gv.pvo && tag.voice != p_gv.pvo)
							continue
						break
					}
				}
//fixme
//				if (i >= gsy.voices.length)
//					??
//console.log('  found i: '+i
// +(p_gv.pst?(' st: '+p_gv.pst):'')
// +(p_gv.pvo?(' vo: '+p_gv.pvo):''))
				if (tag.staff)
					id += '-' + tag.staff
				if (tag.voice)
					id += '-' + tag.voice
				g.part.vo[g.v] = id
				if (!g.v) {
					if (g.part.name)
						id += ' nm="' + g.part.name + '"'
					if (g.part["short-name"])
						id += ' snm="'
							+ g.part["short-name"] + '"'
				}
			} else {
				id = g.part.vo[g.v]
			}
			mus.get_voice(id)
			p_v = mus.get_curvoice()
//console.log('part id '+g.part.id+' '+id+' iv: '+p_v.v)

			// if first sequence, update the staff system
			if (id && !g.mn) {
				csy = mus.get_cur_sy()
				csy.nstaff = gsy.nstaff
				v = p_v.v
				if (!csy.voices[v])
					csy.voices[v] = {}
				p_v.st = p_v.cst = csy.voices[v].st
							= p_gv.st
				if (v > 0 && p_v.st == v_tb[v - 1].st)
					csy.voices[v].second = p_v.second = 1
			}
		} else if (g.part.vo[g.v] == undefined) { // if start of voice
			id = tag.voice || g.part.id || ''
			id += (g.vo + g.v).toString()
			g.part.vo[g.v] = id
			if (!g.v && !g.nm) {	// if first sequence of first voice
				if (g.part.name)
					id += ' nm="' + g.part.name + '"'
				if (g.part["short-name"])
					id += ' snm="' + g.part["short-name"] + '"'
			}
//console.log('seq new v '+id)
			mus.get_voice(id)
			p_v = mus.get_curvoice()
			if (!g.v)			// if first staff of the part
				g.part.st = p_v.st
			if (!p_v.last_sym) {		// if new voice
				csy = mus.get_cur_sy()
				p_v.st = p_v.cst = tag.staff
						? (g.part.st + +tag.staff - 1)
						: g.part.st
				csy.voices[p_v.v].st = p_v.st
				csy.nstaff = g.part.st + g.part.nst - 1
				if (v > 0 && p_v.st == v_tb[v - 1].st)
					csy.voices[v].second = p_v.second = 1
			}
		} else {
			mus.get_voice(g.part.vo[g.v])
			p_v = mus.get_curvoice()
		}

		s = p_v.last_sym
		if (m.end_st			// if ending start
		 && s && s.bar_type) {
			s.text = m.end_st.number
			s.rbstart = 2
		}

		if (g.p && g.p.key)
			val = +g.p.key.fifths
		else if (m.key)
			val = +m.key.fifths
		if (val)
			val = k_tb[val + 7]
		if (g.p && g.p.clef) {		// if clef
			t = g.p.clef.octave
			if (t) {
				t = +t
				if (t > 0)
					t = '^' + (t * 7 + 1)
				else
					t = '_' + (-t * 7 + 1)
			} else {
				t = ''
			}
			val += ' clef=' + g.p.clef.sign + g.p.clef.line + t
		}
		if (val)
			mus.get_key(val)

		if (m.meter)
			mus.new_meter(m.meter.signature)
		if (m.tempo) {
			t = m.tempo[0]
//fixme:other location
			if (!t.location || t.location == "0") {
				mus.new_tempo('1/4=' + t.bpm)
//			} else {
//fixme				??
			}
		}
		if (!g.mn && m.repl) {	// first left repeat bar
			s = {
				typ: C.bar,
				istart: tag.ix,
				bar_type: '|:',
				dur: 0,
				multi: 0
			}
			mus.sym_link(s)
		}
//fixme: why would the segno be on an other element than the next note?
		if (m.segno)
			mus.dcn_add('segno')
	}, // sequence

	// inside <event>
	slur: function(tag) {
	    var	ev, not, s2,
		p_v = mus.get_curvoice(),
		s = p_v.last_sym,
		t = tag["start-note"],
		sl = {
			ty: t.side			// slur type
				? (t.side == "up"
					? C.SL_ABOVE
					: C.SL_BELOW)
				: C.SL_AUTO,
			ss: s				// start symbol (event)
		}

		if (t) {			// starting node
			if (g.nid[t]) {	// if the starting note is already defined
				sl.nts = g.nid[t]	// note
			} else {
				if (!g.sln)
					g.sln = {}
				if (!g.sln[t])
					g.sln[t] = []
				g.sln[t].push(sl)
			}
		}

		t = tag.target
		if (t) {			// the target cannot be defined yet

			// put the slur in the symbol
			if (!s.sls)
				s.sls = []
			s.sls.push(sl)

			if (!g.sl)
				g.sl = {}
			if (!g.sl[t])
				g.sl[t] = []
			g.sl[t].push(sl)

			t = tag["end-note"]
			if (t) {		// the ending note cannot be defined yet
				if (!g.sln)
					g.sln = {}
				if (!g.sln[t])
					g.sln[t] = []
				g.sln[t].push(sl)
			}
			return
		}

		// no target
		switch (tag.location) {
		default:
//		case "outgoing":
			sl.loc = "o"
			if (!s.sls)
				s.sls = []
			s.sls.push(sl)
			break
		case "incoming":
			for (s2 = s.prev; s2; s2 = s2.prev) {
//fixme: may be anything else...
				if (s2.bar_type
				 && s2.bar_type[0] == ':'
				 && s2.text) {
					if (!s2.sls)
						s2.sls = []
					s2.sls.push(sl)
					sl.ss = s2
					sl.se = s
					sl.loc = "i"
					break
				}
			}
			break
		}
	}, // slur

	// inside <system-layout> or <group-layouts>
	"staff-layout": function(tag) {
		if (tag.label)
			g.sy.label = tag.label
// tag
//	label	 label on the left of the staff
//	labelref left of the staff = "name" or "short-name" of the part(voice)
//	symbol	 "brace", "bracket", "line", "none" (def), "square"
	}, // staff-layout

	// inside <page>
	// this function is called when the score has been selected
	// (same function for <system-layout-change>)
	// O tag.layout	ID of the system-layout
	system: function(tag) {
		if (!tag.layout)
			return		// ??
	    var	gsy = g.sysl[tag.layout]

		if (!gsy) {
			mus.error(1, null, "Bad system layout")
			return
		}
		if (tag.measure)
			gsy.loc = tag.measure + ":0"
		else if (tag.location)
			gsy.loc = tag.location
//console.log('system nst '+gsy.nstaff+' '+gsy.staves[gsy.nstaff])
		if (!g.syst)
			g.syst = []
		g.syst.push(gsy)		// chosen staff system
	}, // system

	// inside <layouts>
	// M tag.id
	"system-layout": function(tag) {
		g.sy = {
			staves: [],	// staff system flags
			voices: [],	// st, part, pst, pvo
			nstaff: 0
		}
		g.sysl[tag.id] = g.sy
	}, // system-layout

	// inside <system>
	// tag.layout	ID of the system-layout
	// tag.location (use loc2tim)
	"system-layout-change": function(tag) {
	}, // system-layout-change

	// inside <note>
	tied: function(tag) {
	    var	p_v = mus.get_curvoice(),
		s = p_v.last_sym,
		not = s.notes[s.nhd]

		not.s = s

		switch (tag.location) {
		case "incoming":
			s.ti2 = 1		// end of tie
			not.tie_ty = C.SL_AUTO
			not.tie_s = not		// !!
			break
		case "outgoing":
			s.ti1 = 1		// start of tie
			not.tie_ty = C.SL_AUTO
			break
		default:			// normal case
//fixme: other location?
			if (tag.target) {
				if (!g.ti)
					g.ti = {}
				g.ti[tag.target] = not
//			} else {
//				mus.error
			}
			break
		}
	}, // tied

	// inside <directions-global>
	tempo: function(tag) {
	    var	m = g.meas[g.meas.length - 1]

		if (!m.tempo)
			m.tempo = []
		m.tempo.push(tag)
	}, // tempo

	// inside <directions-global>
	time: function(tag) {
		g.meas[g.meas.length - 1].meter = tag
	}, // time

	// inside <sequence>
	tuplet: function(tag) {
		g.tpl = {
			p: tag.inner[0],
			q: tag.outer[0],
			r: 0,
			f: cfmt.tuplets
		}
	}, // tuplet

	// inside <staff-layout>
	// O tag.label
	// O tag.labelref
	// O tag.stem (part/up/down/float)
	"voice-layout": function(tag) {
	} // voice-layout
    }, // fns

	// --- functions associated to the end of MMX tags
    fne = {

	// end
	beam: function(tag) {
		g.bml--
	}, //beam end

//	// end
//	directions: function(tag) {
//	}, // directions end

	// end
	event: function(tag) {
	    var ev, s1, sl,
		p_v = mus.get_curvoice(),
		s = p_v.last_sym

		if (tag.id) {
			ev = g.bm[tag.id]		// beam level
			switch (ev) {
			case 0:				// stop previous beam
				while (1) {
					s = s.prev
					if (!s)
						break
					if (s.dur) {
						s.beam_end = 1
						break
					}
				}
			case 1:
				s.beam_end = 1
				delete g.bm[tag.id]
				break
			case 2:
				mus.dcn_add('beambr1')
				delete g.bm[tag.id]
				break
			case 3:
				mus.dcn_add('beambr2')
				delete g.bm[tag.id]
				break
			}

			ev = g.sl && g.sl[tag.id]
			if (ev)				// remove the slur pointer
				delete g.sl[tag.id]
		}
	}, // event end

//	// end
//	global: function(tag) {
//	}, // global end

	// end
	grace: function(tag) {
	    var	p_v = mus.get_curvoice(),
		s = g.grace

		// put the grace notes in the grace container
		s.grasym = s.next
		s.next = null
		p_v.last_sym = s
		delete g.grace
	}, // grace

	// end
	"group-layout": function(tag) {
	    var	fl = syfl['c' + g.grll + '_' + tag.symbol],
		st = g.sy.nstaff - 1

//console.log('end group nst '+st+' '+tag.symbol+' '+fl)
		if (fl)
			g.sy.staves[st] |= fl
		g.grll--		// level
	}, // group-layout end

//	// end
//	layouts: function(tag) {
//	}, // layouts end
//
	// end
	measure: function(tag) {
	    var	s, i,
		m = g.meas[g.mn],	// measure global definition
		bar = m.barline		// (from <measure-global>)

		// insert the bar that ends the measure		
		if (bar)
			bar = k_bar[bar]
		if (!bar)
			bar = '|'
		if (m.repr)
			bar = ':|]'

		// in all voices of the part
		for (i = 0; i < g.part.vo.length; i++) {
			mus.get_voice(g.part.vo[i])
			s = {			// s = bar at end of measure
				typ: C.bar,
				istart: tag.ix,
				bar_type: bar,
				dur: 0,
				multi: 0
			}
			if (g.bml)		// if beam crossing the barline
				s.beamon = 1
			switch (m.barline) {
			case "short":
				mus.dcn_add("short")
				break
			case "tick":
				mus.dcn_add("tick")
				break
			}

			mus.sym_link(s)
//console.log('bar V:'+s.p_v.id+' iv: '+s.p_v.v)

			// global definitions
			if (!s.v) {		// if the master voice (1st one)

//fixme: assume location="1"
				if (m.end_end) {
					switch (m.end_end.type) {
					case "stop":
						s.rbstop = 2
						break
					case "discontinue":
						s.rbstop = 1
						break
					}
				}
//fixme: assume location="1"
				if (m.fine)
					mus.dcn_add("fine")
//fixme: assume location="1"
				if (m.jump) {
					switch (m.jump.type) {
					case "segno":
						mus.dcn_add("D.S.")
						break
					case "dsalfine":
						mus.dcn_add("D.S.alfine")
						break
					}
				}
				mus.deco_cnv(s)
			}
		}
	}, // measure end

	// end
	mnx: function(tag) {
	    var	i, gsy,
		csy = mus.get_cur_sy()

		// update the staff flags of the staff system
		// (they were destroyed on new voice)
		if (g.syst) {
			gsy = g.syst[0]
			for (i = 0; i < csy.staves.length; i++)
				csy.staves[i].flags = gsy.staves[i]
		}
//console.log('gene system topv: '+csy.top_voice+' nst: '+csy.nstaff)
//for(var i=0;i<=csy.nstaff;i++)
// console.log(' st '+i+(csy.staves[i].flags!=undefined
//	? (' fl: 0x'+csy.staves[i].flags.toString(16)) :''))
//for(i=0;i<csy.voices.length;i++)
// console.log(' vo '+i+' r: '+csy.voices[i].range+' st: '+csy.voices[i].st)
		mus.generate()
		mus.init_tune()
	}, // mnx end

	// end
	part: function(tag) {
	    var	s, p_v, i,
		m = g.meas[g.mn]	// global definition

		// change the type of the last bars
		if (!m.barline) {
			for (i = 0; i < g.part.vo.length; i++) {
				mus.get_voice(g.part.vo[i])
				p_v = mus.get_curvoice()
				s = p_v.last_sym
				if (s.bar_type == '|')
					s.bar_type = '|]'
			}
		}
		g.vo++			// ready for the next part
	}, // part end

//	// end
//	score: function(tag) {
//	}, // score end
//
	// end
	sequence: function(tag) {
		g.v++			// ready for next voice
	}, // sequence end

	// end
	"staff-layout": function(tag) {
		g.sy.staves[++g.sy.nstaff] = 0	// staff for next <staff-layout>
	}, // staff-layout end

	// end
	"system-layout": function(tag) {
		g.sy.staves.pop()		// remove the unused staff
		g.sy.nstaff--
//console.log('end sl nst '+g.sy.nstaff+' '+g.sy.staves[g.sy.nstaff])
	}, // system-layout end

	// end
	tuplet: function(tag) {
		mus.tp_adj(g.tpl.s, g.tpl.q / g.tpl.p)
		g.tpl = null
	}, // tuplet end

    } // fne

	// common functions
	fns.dim = fns.cresc	// inside <directions>
	fns.dynamics = fns.cresc
	fns.expressions = fns.cresc	// ("subito")
	fns.instructions = fns.cresc	// ("Molto adagio")
	fns["octave-shift"] = fns.cresc
	fns["system-layout-change"] = fns.system
	fns.wedge = fns.cresc

	// add an end of decoration
	function de_add(loc, de) {
	    var	i, s, iloc,
		p_v = mus.get_curvoice()

		if (loc[0] == '#') {		// by event ID
			if (!g.dee)
				g.dee = {}
			loc = loc.slice(1)
			if (!g.dee[loc])
				g.dee[loc] = []
			g.dee[loc].push(de)
		} else {
			iloc = loc2tim(loc)	// [ #measure, time ]
			if (!iloc) {
				mus.error(1, null, "Bad location " + loc)
				return
			}
			if (!p_v.de)
				p_v.de = []
			p_v.de.push([iloc[0], iloc[1], de])
		}
	} // de_add()

	// convert a duration
	function durcnv(dur) {
	    var	r,
		p_v = mus.get_curvoice(),
		d = 1536		// whole note (C.BLEN)

		if (!dur)
			dur = p_v.wmeasure
		else if (!dur.match(/([*/])(\d+)(d*)/))
			return

		r = dur.match(/([*/])(\d+)(d*)/)
		if (r[1] == '*')
			d *= r[2]
		else
			d /= r[2]
		switch (r[3].length) {
		case 1: d *= 1.5; break
		case 2: d *= 1.75; break
		case 3: d *= 1.875; break
		}
		return d
	} // dur_out()

	// convert a location to a time
	// @loc may be
	// - <measure number> ':' <measure time>
	// - <measure time>  (<= measure duration) as
	//	- a fraction of the duration of a whole note(<n> "/" <d>)
	//	- a floating point multiplier of a whole note duration
	// - not '#' (event id)
	// returned value:
	//	[ measure number, time in the measure ]
	function loc2tim(loc) {
	    var	tim, n,
		r = loc.match(/^(\d+:)?(\d+)([/.])(\d+)$/)

		if (!r)
			return
		n = parseInt(r[1]) || g.mn	// measure number
		tim = r[3] == '/'
			? r[2] / r[4]
			: Number(r[2] + r[3] + r[4])
		if (!tim)
			return
		return [n, tim * 1536]
	} // loc2tim()

	// convert to music
	function parse_mnx(tag) {
		if (fns[tag._nm]) {
			parse.istart = tag.ix	// (compatibility with ABC functions)
			if (fns[tag._nm](tag))
				return		// skip when not used or error
		}
		if (tag.children) {
		    var	i,
			n = tag.children.length

			for (i = 0; i < n; i++)
				parse_mnx(tag.children[i])
		}
		if (fne[tag._nm])
			fne[tag._nm](tag)
	} // parse_mnx()

	// reorder the music elements
	function tag_reorder(tag) {
	    var	i,
		nch = [],
		ch = tag.children[0].children	// root.mnx.children

		// get the layouts, then the score then the rest
		for (i = 0; i < ch.length; i++) {
			if (ch[i]._nm == 'layouts')
				nch.push(ch[i])
		}
		for (i = 0; i < ch.length; i++) {
			if (ch[i]._nm == 'score')
				nch.push(ch[i])
		}
		for (i = 0; i < ch.length; i++) {
			if (ch[i]._nm != 'layouts'
			 && ch[i]._nm != 'score')
				nch.push(ch[i])
		}
		tag.children[0].children = nch
		return tag.children[0]		// return <mnx>
	} // tag_reorder()

	// --- main code of mnx2mus() ---

	cfmt.graceslurs = false
	mus.glovar().ulen = C.BLEN / 4
//	mus.param_set_font("dynamfont", "serifItalicBold 16")

	parse.mb = mnx

	// convert to internal music representation
	parse_mnx(tag_reorder(xml2tree(mus, mnx, bof, eof)))
	parse.state = 0		// end of tune
    } // mnx2mus()

} // mnx

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.mnx = tm.mnx
