// pedline.js - module to draw pedal lines instead of 'Ped .. *'
//
// Copyright (C) 2020-2023 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%pedline" appears in a ABC source.
//
// Parameters
//	%%pedline 1

if (typeof tm == "undefined")
	tm = {}

tm.pedline = function(mus) {
    function draw_all_deco(of) {
    var	de, i, x, dp, ds,
	a_de = mus.a_de()

	if (!a_de.length)
		return			// no decoration in this music line
	if (mus.cfmt().pedline) {
		for (i = 0; i < a_de.length; i++) {
			de = a_de[i]
			if (de.dd.name != "ped)")
				continue
			ds = de.start
			dp = de.prev
// ( .. ) ( .. )
//		\ de
//	  \ de.start
//	\ de.prev
// \ de.prev.start
// |_____/\____|
			if (dp && dp.dd.name == "ped)"
			 && dp.s.v == ds.s.v) {
				de.defl.nost =		// /\
					dp.defl.noen = 2
				de.x = ds.s.x - 10
				de.val = de.s.x - ds.s.x - 3
				dp.val = de.x - dp.x

				if (de.y > dp.y)
					de.y = dp.y
				dp.y = de.y
			} else {
				de.x = ds.s.x - 8
				if (!de.defl.noen)
					de.val = de.s.x - ds.s.x - de.s.wl
			}
		}
	}
	of()
    } // draw_all_deco()

    function out_lped(of, x, y, val, defl) {
	if (!mus.cfmt().pedline) {
		of(x, y, val, defl)
		return
	}
	mus.xypath(x, y + 8)
	if (defl.nost) {
		if (defl.nost == 2) {		// \
			mus.out_svg("l2.5 6")
			val -= 2.5
		} else {
			mus.out_svg("m0 6")
		}
	} else {
		mus.out_svg("v6")
	}
	if (defl.noen) {
		if (defl.noen == 2) {		// /
			val -= 2.5
			mus.out_svg("h" + val.toFixed(1) + 'l2.5 -6')
		} else {
			mus.out_svg("h" + val.toFixed(1))
		}
	} else {
		mus.out_svg("h" + val.toFixed(1) + 'v-6')
	}
	mus.out_svg('"/>\n')
    } // out_lped()

    function set_fmt(of, cmd, param) {
	if (cmd == "pedline")
		mus.cfmt().pedline = mus.get_bool(param)
	else
		of(cmd, param)
    } // set_fmt()

	// set the hooks
	mus.draw_all_deco = draw_all_deco.bind(mus, mus.draw_all_deco)
	mus.out_lped = out_lped.bind(mus, mus.out_lped)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
} // pedline

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.pedline = tm.pedline
