// soloffs.js - module to set the X offset of some elements at start of music line
//
// Copyright (C) 2018-2022 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// Parameters
//	%%soloffs <type>=<offset> [<type>=<offset>]*
//		<type> is one of 'part', 'tempo' or 'space'
//		<offset> is the X offset from the start of staff

if (typeof tm == "undefined")
	tm = {}

tm.soloffs = function(mus) {

    function set_fmt(of, cmd, parm) {
	if (cmd == "soloffs") {
	    var	i, v,
		C = tm.C,
		soloffs = mus.cfmt().soloffs = {}

		parm = parm.split(/\s+/)
		while (parm.length) {
			i = parm.shift().split('=')
			v = Number(i[1])
			if (isNaN(v))
				continue		// error
			switch (i[0]) {
//			case 'bar':
//				soloffs[i[0]] = v
//				break
			case 'part':
			case 'space':
				soloffs[i[0]] = v	// see deco.js
				break
			case 'tempo':
				soloffs[i[0]] = v + 16	// see deco.js
				break
//			default:
//				// error
//				break
			}
		}
		return
	}
	of(cmd, parm)
    } // set_fmt()

    function set_sym_glue(of, width) {
    var	s = mus.get_tsfirst(),
	tim = s.time,
	soloffs = mus.cfmt().soloffs

	of(width)		// compute the x offset of the symbols
	if (!soloffs)
		return
	for ( ; s; s = s.ts_next) {
		if (s.time != tim)
			break
		if (soloffs[s.typ.nm] != undefined)
			s.x = soloffs[s.typ.nm]
	}
    } // set_sym_glue()

	// set the hooks
	mus.set_sym_glue = set_sym_glue.bind(mus, mus.set_sym_glue)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
} // soloffs

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.soloffs = tm.soloffs
