// tropt.js - module to optimize the notes after transposition
//
// Copyright (C) 2022-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This module is loaded when "%%tropt" appears in a music source.

if (typeof tm == "undefined")
	tm = {}

tm.tropt = function(mus) {

    // function called before start of the generation
    function voice_adj(of, last_s) {
	if (last_s) {			// if not first time
		of(last_s)
		return
	}
    var	v, p_v, s, m, nt, p, np, na,
	C = tm.C,
	v_tb = mus.get_v_tb(),
	nv = v_tb.length

	// check if some next note has the same pitch without any accidental
	function ok(s, p) {
	    var	nt, m

		while (s) {
			if (s.bar_type)
				return 1 //true
			if (s.typ == C.note) {
				for (m = 0; m <= s.nhd; m++) {
					nt = s.notes[m]
					if (nt.pit == p)
						return nt.acc
				}
			}
			s = s.next
		}
		return 1 //true
	} // ok()

	for (v = 0; v < nv; v++) {
		p_v = v_tb[v]
		if (!p_v.tropt || !p_v.ckey.k_none || !p_v.tr_sco)
			continue
		for (s = p_v.sym; s; s = s.next) {
			if (s.typ != C.note)
				continue
			for (m = 0; m <= s.nhd; m++) {
				nt = s.notes[m]
				if (nt.tie_s) {		// if end of tie
					nt.pit = nt.tie_s.pit
					continue
				}
				p = (nt.pit + 39) % 7		// A..G
				na = 3			// default for new accidental
				switch (nt.acc) {
				case -1:
					switch (p) {
					case 2:		// C
					case 5:		// F
						break	
					default:
						continue
					}
					np = nt.pit - 1
					break
				case -2:
					switch (p) {
					case 2:		// C
					case 5:		// F
						na = -1
						break
					}
					np = nt.pit - 1
					break
				case 1:
					switch (p) {
					case 1:		// B
					case 4:		// E
						break	
					default:
						continue
					}
					np = nt.pit + 1
					break
				case 2:
					switch (p) {
					case 1:		// B
					case 4:		// E
						na = 1
						break	
					}
					np = nt.pit + 1
					break
				default:
					continue
				}
				if (ok(s, np)) {
					nt.pit = np
					nt.acc = na
				}
			}
		}
	}
	of(last_s)
    } // voice_adj()

    // set the tropt parameter
    function do_pscom(of, text) {
	if (text.indexOf("tropt ") == 0)
		mus.set_v_param("tropt", text.split(/\s+/)[1])
	else
		of(text)
    } // do_pscom()

    function set_fmt(of, cmd, parm) {
	if (cmd == "tropt")
		mus.set_v_param("tropt", parm)
	else
		of(cmd, parm)
    } // set_fmt()

    // set the tropt parameter in the current voice
    function set_vp(of, a) {
    var	curvoice = mus.get_curvoice(),
	i = 0

	while (1) {
		if (!a[i])
			break
		if (a[i++] == "tropt"
		 && a[i++] == '=') {
			curvoice.tropt = mus.get_bool(a[i])
			break
		}
	}
	of(a)
    } // set_vp()

	// set the hooks
	mus.do_pscom = do_pscom.bind(mus, mus.do_pscom)
	mus.set_format = set_fmt.bind(mus, mus.set_format)
	mus.voice_adj = voice_adj.bind(mus, mus.voice_adj)
	mus.set_vp = set_vp.bind(mus, mus.set_vp)
} // tropt

if (!tm.mhooks)
	tm.mhooks = {}
tm.mhooks.tropt = tm.tropt
