// txtmus - xml.js - convert xml to a tree of objects
//
// Copyright (C) 2019-2022 Jean-Francois Moine
//
// This file is part of txtmus-core.
//
// txtmus-core is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus-core is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus-core.  If not, see <http://www.gnu.org/licenses/>.

// convert the xml source into a tree
function xml2tree(mus, xml, bof, eof) {
    var	i, j, item, n, no,
	e = bof,
	stack = [],
	root = {
		_nm: "root"
	},
	o = root

	if (!eof)
		eof = 1000000

	// return a pseudo txtmus symbol for error
	function sym() {
		return {
			fname: mus.parse.fname,
			istart: i
		}
	}

	while (1) {
		i = xml.indexOf('<', e)		// tag start
		if (i < 0 || i >= eof)
			break
		if (i > e + 1) {		// if some content
			item = xml.slice(e + 1, i)
				.replace(/\s*/, '') // remove left spaces
			if (item) {
				if (!o.children)
					o.children = []
				o.children.push({
					_nm: "text",
					content: item
				})
			}
		}
		e = xml.indexOf('>', i)		// tag end
		if (e < 0 || e >= eof) {
			mus.error(1, sym(), "No end of tag")
			return root
		}

		switch (xml[i + 1]) {
		case '?':
			continue		// XML header
		case '!':
			e = xml.indexOf('-->', i) // XML comment
			if (e < 0 || e >= eof)
				return root
			continue
		case '/':
			item = xml.slice(i + 2, e)
			if (item != o._nm) {
				mus.error(1, sym(), 'Found tag </' + item +
					'> instead of </'+o._nm + '>')
				continue
			}
			o = stack.pop()
			if (!o)
				mus.error(1, sym(), 'No start tag "' + item + '"')
			continue
		}

		j = 0
		while (1) {				// take care of key="<...>"
			item = xml.slice(i + 1, e).trim()
			n = 0
			while ((j = item.indexOf('"', j) + 1))
				n++
			if (!(n & 1))
				break
			e = xml.indexOf('>', e + 1)	// tag end
		}

		no = {
			_nm: item.match(/[\w-]+/)[0],
			ix: i				// source index
		}
		if (!o.children)
			o.children = []
		o.children.push(no)
		if (item.slice(-1) != '/') {		// if some content
			stack.push(o)
			o = no
		} else {
			item = item.slice(0, -1).trim()
		}
		item = item.replace(/[^\s]*\s*/, '')
		if (!item)
			continue			// no content
		n = item.match(/=|[^\s"=]+|".*?"/g)	// "
		for (j = 0; j < n.length; j += 3) {
			if (n[j + 1] != '='
			 || !n[j + 2]) {
				mus.error(1, sym(), 'Bad key=value in "' + item + '"')
				break
			}
			no[n[j]] = n[j + 2].slice(1, -1)
		}
	}
	return root
} // xml2tree()
