// follow.js - file included in snd-2.js
//
// This script permits to follow the notes while playing.
//
// Copyright (C) 2015-2022 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.

// init
function follow(mus, user, playconf) {
   var	abcplay, playing,
	tune_lst = [],	// array of [tsfirst, voice_tb, info, cfmt] per tune
			// filled on playback start
	playconf = {
		onend: function() {
			playing = 0 //false
		    },
	// fonction called when a music element starts or stops playing
		onnote: function(i, on) {
		    var	b, x, y, i, e, elts

			if (tm.mu)		// if many tunes with same offsets
				elts = tm.mu.d.getElementsByClassName('_' + i + '_')
			else
				elts = document.getElementsByClassName('_' + i + '_')
			if (!elts || !elts.length)
				return			// no symbol?
			e = elts[0]

			e.style.fillOpacity = on ? 0.4 : 0

			// scroll for the element to be in the screen
			if (on && !window.no_scroll) {	
				b = e.getBoundingClientRect()

				// normal
				if (b.top < 0
				 || b.bottom > window.innerHeight * .8)
					y = b.top - window.innerHeight * .3

				// single line
				if (b.left < 0
				 || b.right > window.innerWidth * .8)
					x = b.left - window.innerWidth * .3
				if (x || y)
					window.scrollBy({
						top: y,
						left: x,
						behavior: (x < 0 || y)
								? 'instant'
								: 'smooth'
					})
			}
		    } // onnote()
	} // playconf

	// create a rectangle over a music element
	user.annotate = function(type, start, stop, x, y, w, h, s) {
		switch (type) {
			default: return
			case 'note':
			case 'rest':
				break
		}

		// create a rectangle
		mus.out_svg('<rect class="musr _' + start + '_" x="')
		mus.out_sxsy(x, '" y="', y)
		mus.out_svg('" width="' + w.toFixed(1)
			+ '" height="' + mus.sh(h).toFixed(1)
			+ '"/>\n')
	} // annotate()

	// handle the click on a music element
	function click(evt) {
		evt.preventDefault()
		if (playing) {			// stop playing
			abcplay.stop()
			return
		}
	    var	i, j,
		svg = evt.target,
		e = svg				// keep the clicked element

		// search if click in a SVG image
		while (svg.tagName != 'svg') {
			svg = svg.parentNode
			if (!svg)
				return
		}
		i = svg.getAttribute('class')
		if (!i)
			return
		i = i.match(/tune(\d+)/)
		if (!i)
			return
		i = i[1]			// tune number

		// initialize the play object
		if (!abcplay) {
			if (typeof AbcPlay == "undefined") { // as snd-2.js is not loaded,
				tm.click = function(){}	// don't come here anymore
				return
			}
			if (mus.cfmt().soundfont)
				playconf.sfu = mus.cfmt().soundfont
			tm.abcplay =			// for external access
				abcplay = AbcPlay(playconf)
		}

		// if not done yet,
		// generate the play data of the tune
		if (!tune_lst[i]) {
			tune_lst[i] = mus.tunes[i]
			abcplay.add(tune_lst[i][0],
					tune_lst[i][1],
					tune_lst[i][3])
		}

		// check if click on a music symbol
		s = tune_lst[i][0]		// first symbol of the tune
		i = e.getAttribute("class")
		if (i)
			i = i.match(/_(\d+)_/)
		if (i) {
			i = i[1]		// symbol offset in the source
			while (s && s.istart != i)
				s = s.ts_next
			if (!s) {		// fixme: error ?!
				alert("play bug: no such symbol in the tune")
				return
			}
		}

		// keep the container of the tune being played
		while (s && !s.fname)
			s = s.ts_next
		if (tm.music) {
			for (i = 1; i < tm.music.length; i++) {
				if (tm.music[i].n == s.fname)
					break
			}
			tm.mu = tm.music[i]		// current container
		}

		playing = 1 //true
		abcplay.play(s, null)
	} // click()

tm.click = click
	
////	// install the click callback on the music elements
//	window.addEventListener("click", click)
} // follow()

// create the style of the rectangles
(function () {
    var	sty = document.createElement("style")
	sty.innerHTML = ".musr {fill: #d00000; fill-opacity: 0; z-index: 15}"
	document.head.appendChild(sty)
})()
