// tmdoc-2.js file to include in html pages with tmcore-2.js
//
// Copyright (C) 2014-2022 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.

if (typeof tm == "undefined")
    var tm = {}

	// accept page formatting
	tm.mus_end = function() {}

	// function to load javascript files
	tm.loadjs = function(fn, relay, onerror) {
	    var s = document.createElement('script')
		if (/:\/\//.test(fn)) {
			s.src = fn		// absolute URL
		} else {
		    var	jsdir = document.currentScript ?
			    document.currentScript.src.match(/.*\//) :
			    (function() {
			     var s_a = document.getElementsByTagName('script')
				for (var k = 0; k < s_a.length; k++) {
					if (s_a[k].src.indexOf('tmdoc-') >= 0)
						return s_a[k].src.match(/.*\//) || ''
				}
				return ""	// ??
			})()
			s.src = jsdir + fn
		}
		s.onload = relay
		s.onerror = function() {
			if (onerror)
				onerror(fn)
			else
				alert('error loading ' + fn)
		}
		document.head.appendChild(s)
	} // loadjs()

    var user

(function(){

window.onerror = function(msg, url, line) {
	if (typeof msg == 'string')
		alert("window error: " + msg +
			"\nURL: " + url +
			"\nLine: " + line)
	else if (typeof msg == 'object')
		alert("window error: " + msg.type + ' ' + msg.target.src)
	else
		alert("window error: " + msg)
	return false
}

// function called when tmcore is fully loaded
function tmdoc() {
    var	errtxt = '',
	new_page = '',
	page,				// document source
    user = {	// -- txtmus init argument
	errmsg: function(msg, l, c) {	// get the errors
		errtxt += clean_txt(msg) + '\n'
	},
	img_out: function(str) {	// image output
		new_page += str
	}
    }

// replace <>& by XML character references
function clean_txt(txt) {
	return txt.replace(/<|>|&.*?;|&/g, function(c) {
		switch (c) {
		case '<': return "&lt;"
		case '>': return "&gt;"
		case '&': return "&amp;"
		}
		return c
	})
} // clean_txt()

    // search the ABC tunes and add their rendering as SVG images
    function render() {
    var	i = 0, j, k, res,
	re_stop = /\n<|\n%.begin/g,	// stop on "<" and skip "%%begin"
	mus = new tm.Mus(user),
	re = /(^|\n)(<script[^t>]*type="text\/vnd.abc"|<mei |<mnx>)/g

	for (;;) {

		// get the start of a music sequence
		res = re.exec(page)
		if (!res)
			break
		j = re.lastIndex - res[0].length
		new_page += page.slice(i, j)

		if (page[j + 2] == 's') {
			j = page.indexOf(">", re.lastIndex) + 2
			k = page.indexOf("</script>", j)
		} else if (page[j + 3] == 'e') {
			k = page.indexOf("</mei>", j) + 6
		} else {
			k = page.indexOf("</mnx>", j) + 6
		}
		tune = page.slice(j, k)

		new_page += '<pre style="display:inline-block; vertical-align: top">' +
				clean_txt(tune) +
				'</pre>\n\
<div style="display:inline-block; vertical-align: top">\n'
// not "float:right"
		try {
			mus.tosvg('tmdoc', tune)
		} catch (e) {
			alert("txtmus javascript error: " + e.message +
				"\nStack:\n" + e.stack)
		}
		if (errtxt) {
			i = page.indexOf("\n", j)
			i = page.indexOf("\n", i + 1)
			alert("Errors in\n" +
				page.slice(j, i) +
				"\n...\n\n" + errtxt)
			errtxt = ""
		}

		tm.mus_end()		// close the page if %%pageheight
		new_page += '</div><br/>\n'
		i = page.indexOf('\n', k)
		if (k >= page.length)
			break
		re.lastIndex = i
	}

	try {
		document.body.innerHTML = new_page + page.slice(i)
	} catch (e) {
		alert("txtmus bad generated SVG: " + e.message +
			"\nStack:\n" + e.stack)
	}
    } // render()

	// --- tmdoc() main code ---

	// get the page content
	page = document.body.innerHTML

	// load the required modules, then render the music
	if (tm.modules.load(page, render))
		render()
} // tmdoc()

// loop until tmcore is fully loaded
function dom_loaded() {

	// load the txtmus core if not done by <script>
	if (!tm.Mus) {
		tm.loadjs("tmcore-2.js", dom_loaded)
		return
	}

	window.removeEventListener("load", dom_loaded)

	tmdoc()
}

// wait for the page to be loaded
document.addEventListener("DOMContentLoaded", dom_loaded, false)
})()
