// tmweb-2.js file to include in html pages
//
// Copyright (C) 2014-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.

// This script may be used in HTML or XHTML files.
// It replaces the ABC sequences contained in <script> elements
// with the type "text/vnd.abc" by music as SVG images.
// The other elements stay in place.
// The txtmus core script is automatically loaded.
//
// When the file is .html, if the ABC sequence is contained inside
// elements <script type="text/vnd.abc">, there is no constraint
// about the ABC characters. Note that the <script> tag is removed.
// When the file is .xhtml, if the ABC sequence contains the characters
// '<', '>' or '&', this sequence must be enclosed in a XML comment
// (%<!-- .. %-->) or in a CDATA (%<![CDATA[ .. %]]>).
//
// ABC parameters may be defined in the query string of the URL
// and/or in custon attributes in CSS's.

window.onerror = function(msg, url, line) {
	if (typeof msg == "string")
		alert("window error: " + msg +
			"\nURL: " + url +
			"\nLine: " + line)
	else if (typeof msg == "object")
		alert("window error: " + msg.type + " " + msg.target.src)
	else
		alert("window error: " + msg)
	return false
}

    var	user

if (typeof tm == "undefined")
    var tm = {}

	tm.mu = ""			// container of the current played tune

	tm.mus_end = function() {}	// accept page formatting

	// get the path to the txtmus scripts
	tm.jsdir = document.currentScript
		? document.currentScript.src.match(/.*\//)
		: function() {
		    var s_a = document.getElementsByTagName("script")
			for (var k = 0; k < s_a.length; k++) {
				if (s_a[k].src.indexOf("tmweb") >= 0)
					return s_a[k].src.match(/.*\//) || ""
			}
			return ""	// ??
		}()

	// function to load javascript files
	tm.loadjs = function(fn, relay, onerror) {
	    var	s = document.createElement('script')
		if (/:\/\//.test(fn))
			s.src = fn		// absolute URL
		else
			s.src = tm.jsdir + fn
		s.onload = relay
		s.onerror = function() {
			if (onerror)
				onerror(fn)
			else
				alert('error loading ' + fn)
		}
		document.head.appendChild(s)
	} // loadjs()

// function called when tmweb-2.js is fully loaded
function win_loaded() {
    var	mus, src, outb, err,
	a_inc = {},		// files to be included
	html,			// set if some pure HTML in the file
	busy

//fixme: IntersectionObserver() does not exist in old browsers
//	// set a call back when an element appears on the screen
//	function onVisible(elt, cb) {
//		new IntersectionObserver((ent, obs) => {
//			ent.forEach(entry => {
//				if (entry.intersectionRatio > 0) {
//					cb(elt)
//					obs.disconnect()
//				}
//			})
//		}).observe(elt)
//	} // onVisible()

	// check which <div> became visible
	// if yes, generate it
	function visible() {
	    var	mu, r,
		wh = window.innerHeight || document.documentElement.clientHeight

		while (1) {
			mu = tm.alldiv[0]
			if (!mu)
				break
			r = mu.d.getBoundingClientRect()
			if (r.top > wh)
				break
			musgen(mu)
			tm.alldiv.shift()
		}
		if (tm.alldiv.length) {
			if (!tm.onscroll) {
				tm.onscroll = visible
				window.addEventListener("scroll", visible)
			}
		} else {
			window.removeEventListener("scroll", visible)
		}
	} // visible()
	
	// get the custom properties of a HTML element
	// from 'class="<class>"' and from 'style="--<key>:<value>"'
	function get_p(e) {
	    var	i, j, k, r,
		o = '',
		sh = document.styleSheets,
		s = e.style

		// from class=".."
		c = e.getAttribute("class")
		if (c) {
			c = '.' + c				// selector
			for (i = 0; i < sh.length; i++) {
				r = sh[i].rules
				for (j = 0; j < r.length; j++) {
					if (r[j].selectorText == c)
						break
				}
				if (j < r.length)
					break
			}
			if (i < sh.length) {
				r = r[j]			// rule
				for (i = 0; i < r.style.length; i++) {
					k = r.style[i]
					if (k[0] == '-' && k[1] == '-')
						o += '%%' + k.slice(2) + ' '
							+ r.style.getPropertyValue(k)
							+ '\n'
				}
			}
		}

		// from style=".."
		for (i = 0; i < s.length; i++) {
			k = s[i]
			if (k[0] == '-' && k[1] == '-')
				o += '%%' + k.slice(2) + ' '
					+ s.getPropertyValue(k)
					+ '\n'
		}
		return o
	} // get_p()

	// move the musique sequences to <script> text/vnd.abc
	//	<mei> .. </mei>
	//	<mnx .. > .. </mnx>
	//	<anytag .. class="abc" .. > ..ABC.. </anytag>
	//	%abc-n ..ABC.. '<' with skip %%beginxxx .. %%endxxx
	//	X:n ..ABC.. '<' with skip %%beginxxx .. %%endxxx
	function move_music(src) {
	    var	re, res, i, j, k, t,
		re_stop = /\n<|\n%.begin[^\s]+/g,
		ss = 0,
		out = ""

		if (/<[^>]* class="[^"]*abc[^"]*/.test(src))	// "
			re = '<[^>]* class="[^"]*abc[^"]*'
		else
			re = '%abc-\\d|X:\\s*\\d'
		re = new RegExp('(^|\n)(' + re + ')', 'g')

		while (1) {
			res = re.exec(src)
			if (!res)
				break
			i = re.lastIndex - res[0].length // end of HTML
			if (i > ss) {
				out += src.slice(ss, i)	// copy HTML
				html = 1
			}
			t = res[2]
			if (t[0] == '<') {
				i = src.indexOf('>', i) + 1 // start of ABC
				j = res[2].indexOf(' ')	// get the tag
				t = res[2].slice(1, j)
				j = src.indexOf('</' + t + '>', i)
				ss = j + t.length + 4
			} else {		// %abc or X:

				// get the end of the ABC sequence
				// including the %%beginxxx/%%endxxx sequences
				//	(that may contain XML)
				re_stop.lastIndex = i
				while (1) {
					res = re_stop.exec(src)
					if (!res || res[0] == "\n<")
						break
					k = src.indexOf(res[0].replace("begin", "end"),
							re_stop.lastIndex)
					if (k < 0)
						break
					re_stop.lastIndex = k
				}
				if (!res || k < 0)
					j = src.length
				else
					j = re_stop.lastIndex - 1
				ss = j
			}
			out += '<script type="text/vnd.abc">\n'
				+ src.slice(i, j)
				+ '</script>\n'
			re.lastIndex = ss		// new HTML start
		}
		out += src.slice(ss)		// copy the last HTML sequence
		if (tm.page && html)
			out += '\
<pre class="nop" style="background:#ff8080">\
Printing may be bad because the file contains pure HTML and %%pageheight\
</pre>\n'
		document.body.innerHTML = out
	} // move_music()

	// save the music sequences and replace them by empty <div>'s
	function save_music() {
	    var i, k, div, c, s, sa

		// save MEI or MNX sequences
		function save_spe(ty) {
		    var	s, div

			while (1) {
				s = document.getElementsByTagName(ty)[0]
				if (!s)
					break
				div = document.createElement('div')
				tm.music.push({
					n: 'mus' + tm.music.length,
					t: s.outerHTML,
					d: div
				})
				s.parentNode.replaceChild(div, s)
			}
		} // save_spe()

		// get the global parameters
		tm.music = [{
			t: "",			// global parameters
			n: "mus0"
		}]
		if (a_inc["default.abc"])
			tm.music[0].t = a_inc["default.abc"]

		// get the parameters from the query string of the URL
		k = location.search
		if (k) {
			k = k.substr(1).split("&")
			for (i = 0; i < k.length; i++)
				tm.music[0].t += "%%"
					+ decodeURIComponent(k[i].replace('=', ' '))
					+ '\n'
		}

		// search the <script>'s of the HTML page
		// and replace the music <script>'s by empty <div>'s
		while (1) {
			sa = document.getElementsByTagName('script')
			for (i = 0; i < sa.length; i++) {
				s = sa[i]
				if (s.type == 'text/vnd.abc')
					break
			}
			if (i >= sa.length)
				break

			c = get_p(s)			// get custom properties
			div = document.createElement('div')
			if (!tm.music[0].t
			 && s.text.indexOf('\nX:') < 0) {
				tm.music[0].t = c + s.innerHTML // global
				if (!tm.music[0].d)
					tm.music[0].d = div
			} else {
				tm.music.push({
					n: 'mus' + tm.music.length,
					t: c + s.innerHTML,
					d: div
				})
			}
			s.parentNode.replaceChild(div, s)
		}

		// special treatment for MEI and MNX
		save_spe('mei')
		save_spe('mnx')
	} // save_music()

	// generate a music sequence
	// @mu = object { t: music source, d: <div> }
	function musgen(mu) {
	    var	t = mu.t

		if (busy) {
			mu.w = 1 //true
			return
		}
		busy = 1 //true

		// render a music sequence
		function render() {
		    var	i, j, e, o,
			f = mus.cfmt().with_source	// options

			// start the generation
			outb = err = ""
			if (t[0] != '<')
				mus.tosvg(mu.n, t)
			else if (t[2] == 'e')
				tm.mei.mei2mus.call(mus, mu.n, t)
			else
				tm.mnx.mnx2mus.call(mus, mu.n, t)

			tm.mus_end()		// close the page if %%pageheight

			// mu.d can be null when parameters in query string
			// and no ABC script with global parameters
			if (mu.d) {

				if (err)
					outb +=
					    '<pre class="nop" style="background:#ff8080">'
						+ err + "</pre>\n"
				if (f && outb) {
					o = outb		// SVG images and errors
					outb = '<pre class="source pre"'
					if (f.indexOf('edit'))
						outb += " contentEditable"
					outb += '>'
						+ clean_txt(t)
						+ '</pre>\n\
<div class="source">\n'
						+ o
						+ '</div>\n'
				}

				mu.d.innerHTML = outb	// update the browser

				mu.d.addEventListener('click', tm.click)
				e = mu.d.getElementsByTagName('svg')
				for (i = 0; i < e.length; i++) {
					j = e[i].getAttribute('class')
					if (j)			// (if no page formatting)
						j = j.match(/tune(\d+)/)
					if (!j)
						continue
					j = j[1]		// tune number
//					tune_lst[j] = null	// get new play references
				}
			}

			// if some generation waiting, start it
			mu.w = busy = 0 //false
			for (i = 1; i < tm.music.length; i++) {
				if (tm.music[i].w) {
					musgen(tm.music[i])
					break
				}
			}
		} // render()

		// load the possible %%abc-include files
		function include() {
		    var	i, j, fn, r,
			k = 0

			while (1) {
				i = t.indexOf('%%abc-include ', k)
				if (i < 0) {
					render()
					return
				}
				i += 14
				j = t.indexOf('\n', i)
				fn = t.slice(i, j).trim()
				if (!a_inc[fn])
					break
				k = j
			}

			// %%abc-include found: load the file
			r = new XMLHttpRequest()
			r.open('GET', fn, true)		// (async)
			r.onload = function() {
				if (r.status === 200) {
					a_inc[fn] = r.responseText
					if (tm.modules.load(a_inc[fn], include))
						include()
				} else {
					a_inc[fn] = '%\n'
					alert('Error getting ' + fn + '\n' + r.statusText)
					include()
				}
			}
			r.onerror = function () {
				a_inc[fn] = '%\n'
				alert('Error getting ' + fn + '\n' + r.statusText)
				include()
			}
			r.send()
		} // include()

		// load the required modules and the include files before rendering
		if (tm.modules.load(t, include))
			include()
	} // musgen()

	tm.musgen = musgen			// used in set_music() below

// replace <>& by XML character references
	function clean_txt(txt) {
		return txt.replace(/<|>|&.*?;|&/g, function(c) {
			switch (c) {
			case '<': return "&lt;"
			case '>': return "&gt;"
			case '&': return "&amp;"
			}
			return c
		})
	} // clean_txt()

	// --- win_loaded() main code ---

	// load the txtmus core if not done by <script>
	if (!tm.Mus) {
		tm.loadjs("tmcore-2.js", win_loaded)
		return
	}

	// initialize the user object
	user = {
		read_file: function(fn) {
			return a_inc[fn]
		},
		errmsg: function(msg, l, c) {	// get the errors
			err += clean_txt(msg) + '\n'
		},
		img_out: function(p) {		// image output
			outb += p
		}
	} // user

	src = document.body.innerHTML
	if (src.indexOf('type="text/vnd.abc"') < 0)
		move_music(src)

	save_music()	// save the music sequences and replace them by <div>

	// initialize the generation
	tm.mus = 				// for external access
		mus = new tm.Mus(user)
	if (typeof follow == "function")	// if snd-2.js loaded
		follow(mus, user)		// initialize the play follow
	if (tm.music[0].t)
		musgen(tm.music[0])

	// create a list of all <div>'s
	tm.alldiv = []
	for (var i = 1; i < tm.music.length; i++)
		tm.alldiv.push(tm.music[i])
	visible()
} // win_loaded()

// ---- interface for changing the music source and redisplaying it ----
// get the music source of a <div> element
tm.get_music = function(d) {
    var	i, mu

	for (i = 1; i < tm.music.length; i++) {
		mu = tm.music[i]
		if (mu.d == d)
			return mu.t
	}
//	return ""
} // get_music()

// set the music source of a <div> element and redisplay it
tm.set_music = function(d, t) {
    var	i, mu

	for (i = 1; i < tm.music.length; i++) {
		mu = tm.music[i]
		if (mu.d == d) {
			mu.t = t
			tm.musgen(mu)
			break
		}
	}
} // set_music()

// wait for the scripts to be loaded
window.addEventListener("load", win_loaded, {once:true})
