// tmweb1-2.js file to include in html pages for ABC rendering
//
// Copyright (C) 2019-2024 Jean-Francois Moine
//
// This file is part of txtmus.
//
// txtmus is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// txtmus is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with txtmus.  If not, see <http://www.gnu.org/licenses/>.
//
// This script either:
// - builds a list of the tunes when there is no selection or
// - displays the selected tune.
//
// When the tune is displayed, if playing is not enabled,
// scrolling the music may be done by clicking/taping
// on the 'start scrolling' button.
//
// The header of the tune list ("Tunes:") may be set in a global
// javascript variable 'list_head'.
// The tail of the tune list ("(all tunes)") may be set in a global
// javascript variable 'list_tail'.

window.onerror = function(msg, url, line) {
	if (typeof msg == 'string')
		alert("window error: " + msg +
			"\nURL: " + url +
			"\nLine: " + line)
	else if (typeof msg == 'object')
		alert("window error: " + msg.type + ' ' + msg.target.src)
	else
		alert("window error: " + msg)
	return false
}

// remove the menu button on print
window.onbeforeprint = function() {
   var	e = document.getElementById("dd")
	if (e)
		e.style.display = "none"
}
window.onafterprint = function() {
   var	e = document.getElementById("dd")
	if (e)
		e.style.display = "block"
}

    var user
if (typeof tm == "undefined")
    var tm = {}

	// accept page formatting
	tm.mus_end = function() {}
	tm.a_inc = {}		// files to be included

// function called when tmcore is fully loaded
function win_loaded() {
    var	mus,
	new_page,

	tune_dur,			// scroll tune duration
	scroll_to,			// scroll timeout
	dt,				// scroll delta per timeout
	sY,				// current scroll Y

	page,				// document source
	a_inc = tm.a_inc,
	errtxt = '',
	app = "tmweb1",
	jsdir = document.currentScript
		? document.currentScript.src.match(/.*\//)
		: function() {
		    var	s_a = document.getElementsByTagName('script')

			for (var k = 0; k < s_a.length; k++) {
				if (s_a[k].src.indexOf(app) >= 0)
					return s_a[k].src.match(/.*\//) || ''
			}
			return ""	// ??
		}()
// end of the variables of win_loaded()

//fixme: test for log in Android
//	function log(p) {
//		e = document.getElementById("log")
//		if (e)			// (no e when source edit)
//			e.value += p + '\n'
//	}

	// extract the music text from the HTML body
//fixme: only one of ABC, MEI or MNX
	function fix_abc(s) {
	    var	i, j,
		r = s.match(/<mnx|<mei|vnd.abc/)

		if (!r)
			return s
		switch (r[0]) {
		case "<mei":
			i = s.indexOf('<mei')
			j = s.indexOf('</mei>', i)
			return s.slice(i, j + 6)
		case "<mnx":
			i = s.indexOf('<mnx')
			j = s.indexOf('</mnx>', i)
			return s.slice(i, j + 6)
		}

		// ABC
		i = s.indexOf('type="text/vnd.abc"')
		i = s.indexOf('\n', i) + 1
		j = s.indexOf('</script', i)
		return s.slice(i, j)
	} // fix_abc()

	// replace <>& by XML character references
	function clean_txt(txt) {
		return txt.replace(/<|>|&.*?;|&/g, function(c) {
			switch (c) {
			case '<': return "&lt;"
			case '>': return "&gt;"
			case '&': return "&amp;"
			}
			return c
		})
	}

	// scroll the displayed music
	function do_scroll(old) {
	    var	d, ttop

		// get the duration of the (first) tune
		function get_dur() {
		    var	d, i, n, pf,
			s = mus.tunes[0][0]	// first symbol of the first tune

			while (1) {
				if (s.tempo && !pf) {
					d = 0
					n = s.tempo_notes.length
					for (i = 0; i < n; i++)
						d += s.tempo_notes[i]
					pf = d * s.tempo / 60
				}
				if (!s.ts_next)
					break
				s = s.ts_next
			}
			if (!pf)
				pf = tm.C.BLEN / 8	// default: Q:1/4=120
//				     tm.C.BLEN / 4 * 120 / 60
			tune_dur = s.time / pf
		} // get_dur()

		// if start, get the window parameters and delay the first scroll
		if (!old) {
			d = document.documentElement

			if (!tune_dur)
				get_dur()

			// time for scrolling one pixel
			dt = tune_dur / d.scrollHeight

			// start scrolling at this time (1/4 of the first screen)
			ttop = dt * d.clientHeight / 4
			document.getElementById("ss").style.display = "block"
			scroll_to = setTimeout(do_scroll, ttop * 1000, 1)

			// (in Android Browser, remove the address bar)
			window.scrollTo(0, 8)		// go to the top
			sY = 0
		} else {
			if (sY == window.pageYOffset) {	// no scroll -> finished
				document.getElementById("ss").style.display = "none"
				scroll_to = null
				return
			}
			sY = window.pageYOffset
			window.scrollTo(0, sY + 1)
			scroll_to = setTimeout(do_scroll, dt * 1000, 1)
		}
	} // do_scroll()

	// source edit
	tm.src_upd = function() {
		page = document.getElementById('ta').value
		tm.get_sel()
	} // src_upd()

	tm.src_edit = function() {

		// offer a textarea with the ABC source and 2 buttons
		document.body.innerHTML = '\
<textarea id="ta" rows="50" cols="80" style="overflow:scroll">'
			+ page + '</textarea>\
<br/>\
<a href="#" onclick="tm.src_upd()"> Apply </a> - \
<a href="#" onclick="tm.get_sel()"> Cancel </a>'
	} // src_edit()

	// start/stop scrolling when no play
	tm.st_scroll = function() {
		if (scroll_to) {
			clearTimeout(scroll_to)
			document.getElementById("ss").style.display = "none"
			scroll_to = null
		} else {
			scroll_to = setTimeout(do_scroll, 500, 0)	// scroll start
		}
	} // st_scroll()

	// function to load javascript files
	tm.loadjs = function(fn, relay, onerror) {
	    var	s = document.createElement('script')
		if (/:\/\//.test(fn))
			s.src = fn		// absolute URL
		else
			s.src = jsdir + fn
		s.onload = relay
		s.onerror = function() {
			if (onerror)
				onerror(fn)
			else
				alert('error loading ' + fn)
		}
		document.head.appendChild(s)
	} // loadjs()

	// build a list of the tunes
	tm.get_sel = function() {
		if (page[0] == '<')
			return tm.do_render()	// MEI or MNX

	    var	j, k,
		n = 0,
		i = 0,
		t = (typeof list_head == "undefined" ? "Tunes:" : list_head) + '<p>\n'
		tt = typeof list_tail == "undefined" ? "(all tunes)" : list_tail

//fixme: problem if some (or only) MEI/MNX tunes
		for (;;) {
			i = page.indexOf("\nX:", i)
			if (i < 0)
				break
			k = page.indexOf("\n", ++i)
			j = page.indexOf("\nT:", i)
			n++
			t += ' - <a href="#" \
onclick="tm.do_render(\'' + page.slice(i, k) + '$\')">' +
				page.slice(i + 2, k).replace(/%.*/,'')
			if (j > 0 && j < i + 20) {
				k = page.indexOf("\n", j + 1)
				t += " " + page.slice(j + 3, k).replace(/%.*/,'')
				if (page[k + 1] == 'T' && page[k + 2] == ':') {
					j = k + 3
					k = page.indexOf("\n", j)
					t += " - " + page.slice(j, k).replace(/%.*/,'')
				}
			}
			t += '</a><br>\n'
			i = k
		}
		if (n <= 1)
			return tm.do_render()	// only one tune

		t += ' - <a href="#" \
onclick="tm.do_render(\'.*\')">' + tt +
			'</a><br>\n'

		document.body.innerHTML = t
		if (window.location.hash)
			window.location.hash = ''
	} // get_sel()

	// menu click
	function mclick(e) {
		e.preventDefault()
	    var	el = document.getElementById("dc")
		if (e.type == "click")
			el.classList.toggle("show")
		else					// e.type == mouseleave
			el.classList.remove("show")	// remove the menu
	} //mclick()
	// search/ask the tune to be rendered
	function render() {
	    var	select = window.location.hash.slice(1)		// after '#'

		// create styles for the menu
	   var	sty = document.createElement('style')
		sty.innerHTML = '\
.dd{position:fixed;top:0;bottom:0;right:0;height:40px;cursor:pointer;font-size:16px}\
#ss{display:none;background-color:red}\
.db{margin:5px;background-color:yellow}\
.db:hover,.db:focus{background-color:lightgreen}\
.dc{position:absolute;left:-70px;min-width:100px;display:none;background-color:yellow}\
.dc label{display:block;padding:0 5px 0 5px;margin:2px}\
.dc label:hover{outline:solid;outline-width:2px}\
.show{display:block}'
		document.head.appendChild(sty)

		// if no selection and many tunes, get the references of the tunes
		// and ask which one to display
		if (!select)
			tm.get_sel()
		else
			tm.do_render(decodeURIComponent(select))
	} // render()

	// replace the (previous) body by the music
	tm.do_render = function(select) {
//		if (e)
//			e.preventDefault()

		tune_lst = []
		tune_dur = 0
		tm.mus =				// for external access
			mus = new tm.Mus(user)
		new_page = ""

		// initialize the play follow function
		if (typeof follow == "function")
			follow(mus, user)

		if (select) {
			mus.tosvg(app, "%%select " + select)
			window.location.hash = encodeURIComponent(select)
		}
		try {
			mus.tosvg(app, page)
		} catch (e) {
			alert("tm javascript error: " + e.message +
				"\nStack:\n" + e.stack)
		}

		tm.mus_end()		// close the page if %%pageheight
		if (errtxt) {
			new_page += '<pre class="nop" style="background:#ff8080">' +
					errtxt + "</pre>\n"
			errtxt = ""
		}

		// add the menu
		new_page += '\
<div id="dd" class="dd nop">\
<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" class="db">\
<path d="M4 6h15v2.5H4zm0 5h15v2.5H4zm0 5h15v2.5H4z" fill="black"/>\
</svg>\
<div id="dc" class="dc">\
<label id="edit" onclick="tm.src_edit()">Source edit</label>\
<label id="list" onclick="tm.get_sel()">Tune list</label>\
<label id="play" onclick="tm.st_scroll()">Scroll</label>\
</div>\
</div>\
<label id="ss" class="dd nop" onclick="tm.st_scroll()">Scroll<br/>stop</label>'
//fixme: test for log in Android
//new_page += '<textarea id="log" rows="30" cols="60">-- log --\n</textarea>'

		// change the page
		try {
			document.body.innerHTML = new_page
		} catch (e) {
			alert("txtmus bad generated SVG: " + e.message +
				"\nStack:\n" + e.stack)
			return
		}

		// add the event handlers
	    var	elts = document.getElementsByTagName('svg')
		for (var i = 0; i < elts.length; i++) {
			e = elts[i]
			if (e.getAttribute('class').indexOf('tune') >= 0)
				e.addEventListener('click', tm.click)
		}
		e = document.getElementById("dd")
		e.addEventListener('click', mclick)
		e.addEventListener('mouseleave', mclick)

		// update the menu
		setTimeout(function() {

			// remove scroll in the menu if not a big tune
			if (
//			    typeof AbcPlay != "undefined" ||
			    document.documentElement.scrollHeight <= window.innerHeight)
				document.getElementById("play").style.display = "none"
		}, 500)
	} // do_render()

	// load the %%abc-include files
	function include() {
	    var	i, j, fn, r,
		k = 0

		while (1) {
			i = page.indexOf('%%abc-include ', k)
			if (i < 0) {
				render()
				return
			}
			i += 14
			j = page.indexOf('\n', i)
			fn = page.slice(i, j).trim()
			if (!a_inc[fn])
				break
			k = j
		}

		// %%abc-include found: load the file
		r = new XMLHttpRequest()
		r.open('GET', fn, true)		// (async)
		r.onload = function() {
			if (r.status === 200) {
				a_inc[fn] = r.responseText
				if (tm.modules.load(a_inc[fn], include))
					include()
			} else {
				a_inc[fn] = '%\n'
				alert('Error getting ' + fn + '\n' + r.statusText)
				include()
			}
		}
		r.onerror = function () {
			a_inc[fn] = '%\n'
			alert('Error getting ' + fn + '\n' + r.statusText)
			include()
		}
		r.send()
	} // include()

	// --- win_loaded() main code ---

	// load the txtmus core if not done by <script>
	if (!tm.Mus) {
		tm.loadjs("tmcore-2.js", win_loaded)
		return
	}

	// initialize the user object
	user = {
		read_file: function(fn) {
			return a_inc[fn]
		}, // read_file()
		errmsg: function(msg, l, c) {	// get the errors
			errtxt += clean_txt(msg) + '\n'
		},
		img_out: function(str) {	// image output
			new_page += str
		}
	} // user

	// get the page content
	page = fix_abc(document.body.innerHTML)

	// load the required modules and the include files before rendering
	if (tm.modules.load(page, include))
		include()
} // win_loaded()

// wait for the scripts to be loaded
window.addEventListener("load", win_loaded)
